#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all Matplotlib-based plot and animation subclasses.
'''

# ....................{ TODO                               }....................
#FIXME: Visuals currently exhibit extremely non-ideal behaviour, including:
#
#* Repeated display of OS-level notifications resembling '"Figure 1" is ready'.
#* Continually bringing the window containing the most recently created plot or
#  animation frame to the front of the window stack, which then forcefully
#  steals the keyboard focus.
#
#Altogether, these issues effectively render interactive plotting unusable.
#Fortunately, the following StackOverflow post appears to have sufficiently
#resolved this issue for relatively recent versions of Matplotlib:
#    https://stackoverflow.com/a/55456635/2809027
#Likewise, note the related discussion at the close of this issue thread:
#    https://github.com/matplotlib/matplotlib/issues/596
#
#The core issue appears to be our use of the pyplot.pause() function, which we
#will now need to globally replace across the entire codebase with calls to
#fig.canvas.draw_idle() and fig.canvas.start_event_loop(0.001) instead.

#FIXME: Refactor all procedural cell cluster-specific
#"betse.science.visual.plot.plotutil" functions into subclasses of the
#"LayerCellsABC" base class defined elsewhere. Ultimate power fights the dark deceit!

# ....................{ IMPORTS                            }....................
import numpy as np
from abc import ABCMeta
from betse.exceptions import BetseSimVisualException
from betse.lib.matplotlib import mplfigure
from betse.lib.matplotlib.matplotlibs import mpl_config
from betse.lib.matplotlib.mplzorder import ZORDER_PATCH, ZORDER_STREAM
from betse.lib.matplotlib.mplutil import ignoring_deprecations_mpl
from betse.lib.numpy import nparray
from betse.science.config.export.visual.confexpvisabc import (
    SimConfVisualCellsABC)
from betse.science.math import mathunit
from betse.science.phase.phasecls import SimPhase
from betse.util.io.log import logs
from betse.util.py import pyref
from betse.util.type import types
from betse.util.type.iterable import iterget
from betse.util.type.obj import objiter, objtest
from betse.util.type.types import (
    type_check,
    IterableTypes,
    NoneType,
    NumericOrNoneTypes,
    SequenceTypes,
    SequenceOrNoneTypes,
    StrOrNoneTypes,
)
from matplotlib import pyplot
from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from matplotlib.collections import PolyCollection
from matplotlib.colors import Colormap
from matplotlib.figure import Figure
from matplotlib.image import AxesImage
from matplotlib.patches import FancyArrowPatch
from matplotlib.streamplot import StreamplotSet

# ....................{ SUPERCLASSES                       }....................
class VisualCellsABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all **cell cluster visual** (i.e., plot or animation
    spatially visualizing the current cell cluster) subclasses.

    Subclasses of this class plot the spatial distribution of one or more
    modelled variables (e.g., charge distribution, membrane voltage) for
    either:

    * A single simulation time step, in which this subclass plots a still frame
      of these variables for this step.
    * One or more simulation time steps, in which this subclass animates a
      video of these variables for these steps.

    Attributes (Private)
    ----------
    _conf : SimConfVisualCellsABC
        Configuration for this visual, synchronized with the user-defined
        YAML-backed simulation configuration file for this phase.
    _kind : str
        Basename of the subdirectory in the phase-specific results directory
        to which all files exported for this visual are saved *and*
        the basename prefix of these files.
    _layers : list
        List of all :class:`LayerCellsABC` instances collectively composing
        this visual.
    _phase : SimPhase
        Current simulation phase.

    Attributes (Private: Figure)
    ----------
    _figure : Figure
        Matplotlib figure providing the current animation frame.
    _figure_title : str
        Text displayed above the figure itself.

    Attributes (Private: Axes)
    ----------
    _axes : Axes
        Matplotlib figure axes providing the current animation frame data.
    _axes_title : str
        Text displayed above the figure axes. If a non-``None`` value for the
        ``axes_title`` parameter is passed to the :meth:`__init__` method, this
        is that value; else, this is the value of the ``figure_title``
        parameter passed to that method.
    _axes_x_label : str
        Text displayed below the figure's X axis.
    _axes_y_label : str
        Text displayed to the left of the figure's Y axis.

    Attributes (Private: Color)
    ----------
    _color_mappables : IterableTypes
        Iterable of all plotted mappables (i.e., instances of the
        :class:`ScalarMappable` class added to the figure axes of this plot or
        animation) whose minimum and maximum values define the color range
        displayed both in the figure axes and colorbar of this plot or
        animation. Due to matplotlib constraints, only the first mappable in
        this iterable is associated with this colorbar.
    _colorbar_title: StrOrNoneTypes
        Text displayed above the figure colorbar if any *or* ``None``
        otherwise.
    _color_max : NumericSimpleTypes
        Maximum color value to be displayed by the colorbar.
    _color_min : NumericSimpleTypes
        Minimum color value to be displayed by the colorbar.
    _colormap : Colormap
        Matplotlib colormap with which to create this animation's colorbar.

    Attributes (Private: Saving)
    ----------
    _writer_savefig_kwargs : dict
        Dictionary of all keyword arguments to be passed to the
        :meth:`Figure.savefig` method called to save each visual frame for
        both still images and animated video. By default, this dictionary:

        * Enables the ``transparent`` argument, painting frame backgrounds
          transparently rather than as a solid color (usually, white).

    Attributes (Private: Time)
    ----------
    _is_time_step_first : bool
        ``True`` only if the first frame either has yet to be plotter *or* is
        currently being plotted.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,

        # Mandatory parameters.
        phase: SimPhase,
        conf: SimConfVisualCellsABC,
        is_save: bool,
        is_show: bool,
        kind: str,
        figure_title: str,

        # Optional parameters.
        axes_title: StrOrNoneTypes = None,
        axes_x_label: str = 'Spatial Distance [um]',
        axes_y_label: str = 'Spatial Distance [um]',
        colorbar_title: StrOrNoneTypes = None,
        colormap: Colormap | None = None,

        #FIXME: Refactor this to be mandatory instead.
        layers: SequenceOrNoneTypes = None,
    ) -> None:
        '''
        Initialize this visual.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        conf : SimConfVisualCellsABC
            Configuration for this visual, synchronized with the user-defined
            YAML-backed simulation configuration file for this phase.
        is_save : bool
            ``True`` only if non-interactively saving this visual.
        is_show : bool
            ``True`` only if interactively displaying this visual.
        kind : str
            Terse machine-readable string (e.g., ``Vmem``) serving as both:

            * The basename of the subdirectory of the phase-specific results
              directory containing all files saved by this visual.
            * The basename prefix of these files.
        figure_title : str
            Text displayed above the figure itself.
        axes_title : optional[str]
            Text displayed above the figure axes but below the figure title.
            Defaults to `None`, in which case no such text is displayed.
        axes_x_label : optional[str]
            Text displayed below this figure's X axis. Defaults to ``None``, in
            which case a general-purpose string is defaulted to.
        axes_y_label : optional[str]
            Text displayed to the left of this figure's Y axis. Defaults to
            ``None``, in which case a general-purpose string is defaulted to.
        colorbar_title: optional[str]
            Text displayed above the figure colorbar if any *or* ``None``
            otherwise. If this parameter is ``None`` *and* at least one layer
            in the passed ``layers`` parameter is an instance of the
            :class:`LayerCellsColorfulABC` base class, an exception is raised.
        colormap : optional[Colormap]
            Matplotlib colormap to be used by default for all animation artists
            (e.g., colorbar, images). Defaults to ``None``, in which case the
            default colormap is used.
        layers : optional[SequenceTypes]
            Sequence of all :class:`LayerCellsABC` instances collectively
            plotting each frame of this visual. **Order is extremely
            significant.** Specifically, the order of layers in this sequence
            defines the order in which these layers are plotted and hence
            overlaid onto one another (i.e., z-order). Defaults to `None`, in
            which case the subclass is responsible for manually plotting this
            visual.
        '''

        # Default unpassed parameters.
        if colormap is None:
            colormap = phase.p.default_cm
        if layers is None:
            layers = ()

        # Classify *AFTER* validating parameters.
        self._phase = phase
        self._conf = conf
        self._axes_title = axes_title
        self._axes_x_label = axes_x_label
        self._axes_y_label = axes_y_label
        self._colorbar_title = colorbar_title
        self._colormap = colormap
        self._is_save = is_save
        self._is_show = is_show
        self._figure_title = figure_title
        self._kind = kind

        # Convert the passed sequence of layers into a new list of layers,
        # permitting this sequence to be safely modified *WITHOUT* modifying
        # the passed sequence. To validate the type of each such layer, call an
        # existing method rather than manually performing this conversion.
        self._layers = []
        self._append_layer(*layers)

        # If autoscaling colors, ignore the passed minimum and maximum. To
        # avoid propagating these changes back to the underlying configuration
        # file, these changes are isolated to new instance variables; the
        # passed "conf.color_max" and "conf.color_min" variables are unchanged.
        if self._conf.is_color_autoscaled:
            self._color_max = None
            self._color_min = None
        # Else, classify the passed minimum and maximum.
        else:
            self._color_max = conf.color_max
            self._color_min = conf.color_min

        # Default all attributes to be subsequently defined.
        self._color_mappables = None
        self._writer_frames = None
        self._writer_video = None

        # Default all remaining attributes.
        self._is_time_step_first = True

        # Dictionary of keyword arguments to be passed to Figure.savefig().
        self._writer_savefig_kwargs = {
            # Visualize frame backgrounds transparently rather than solid.
            'transparent': True,
        }

        # Initialize this plot's figure *AFTER* defining all attributes.
        self._init_figure()

    # ..................{ INITIALIZERS ~ figure              }..................
    def _init_figure(self) -> None:
        '''
        Initialize this plot's figure.

        If an axes title is defined, the mandatory figure title is added to
        this figure as a "supertitle" and that axes title to this figure axes
        as a "subtitle;" else, this figure title is added to this figure axes.
        Likewise, the colorbar title is added to this figure's colorbar.
        '''

        # Figure encapsulating this visual as a weak rather than strong (the
        # default) reference, avoiding circular references and complications
        # thereof (e.g., memory overhead). Figures created by the "pyplot" API
        # are internally retained in Matplotlib's "Gcf" figure cache until
        # explicitly closed -- either non-interactively by a close() call or
        # interactively by the corresponding GUI window being closed. Hence,
        # strong figure references should typically *NOT* be retained.
        #
        # Note that this is the only remaining "pyplot" function which has no
        # object-oriented analogue in the Matplotlib API. Ideally, we would
        # simply instantiate this figure directly: e.g.,
        #
        #     self._figure = Figure()
        #
        # Sadly, doing so is complicated by the fact that a figure requires a
        # backend-specific canvas, which the low-level
        # pyplot.new_figure_manager() function called by the high-level
        # pyplot.figure() function instantiates here on our behalf. Attempting
        # to manually perform such logic in a forward-compatible manner
        # effectively remains infeasible. So, we necessarily defer to "pyplot".
        self._figure = pyref.proxy_weak(pyplot.figure())
        # self._figure = pyplot.figure()

        # Figure axes scaled to the extent of the current 2D environment as a
        # weak rather than strong (the default) reference, thus avoiding
        # circular references and complications thereof (e.g., memory
        # overhead). Since figures already contain their axes as a strong
        # reference, we need *NOT* do so as well here.
        self._axes = self._figure.add_subplot(111)

        # If this object was initialized with both a figure and axes title,
        # display the former above the latter.
        if self._axes_title:
            # logs.log_debug('Setting supertitle!')
            self._figure.suptitle(
                self._figure_title, fontsize=14, fontweight='bold')
        # Else, display the figure title as the axes title.
        else:
            # logs.log_debug('Setting normal title!')
            self._axes_title = self._figure_title
        assert types.is_str_nonempty(self._axes_title), (
            types.assert_not_str_nonempty(self._axes_title, 'Axis title'))

        # Initialize the X and Y axes of this plot's figure *AFTER* classifying
        # the axes title above.
        self._init_figure_axes()


    def _init_figure_axes(self) -> None:
        '''
        Initialize the X and Y axes of this plot's figure.
        '''

        # Bound these axes by this extent.
        self._axes.axis('equal')
        self._axes.axis(self._phase.cache.upscaled.extent)

        # Display passed human-readable strings as axes attributes.
        self._axes.set_xlabel(self._axes_x_label)
        self._axes.set_ylabel(self._axes_y_label)
        self._axes.set_title(self._axes_title)


    def _reinit_figure_axes(self) -> None:
        '''
        Reinitialize the X and Y axes of this plot's figure, removing all
        artists previously drawn to these axes.

        This method is typically called by animation subclasses _before_
        redrawing cell data of the current animation frame onto these axes.
        '''

        # Remove all artists previously drawn to these axes.
        self._axes.clear()

        # Reinitialize these axes.
        self._init_figure_axes()

    # ..................{ PREPARERS ~ figure                 }..................
    #FIXME: This method should no longer accept any parameters. All parameters
    #currently accepted by this method are obsolete and hence remain
    #undocumented. They should *ALL* be removed as soon as feasible.
    def _prep_figure(self, *args, **kwargs) -> None:
        '''
        Prepare this visual *after* having previously initialized this
        visual but *before* subsequently displaying and/or saving this
        visual.
        '''

        # Prepare all layers to be layered onto this visual *BEFORE*
        # autoscaling colors assuming layers to have been prepared.
        self._prep_layers()

        #FIXME: Obsolete. Remove after adopting layers everywhere.
        if 'color_data' in kwargs:
            self._prep_colors_obsolete(*args, **kwargs)

    # ..................{ DEINITIALIZERS                     }..................
    def close(self) -> None:
        '''
        Deallocate all resources associated with this visual.

        To reduce resource overhead (namely memory), this method (in order):

        #. Explicitly clears this visual's figure axis and figure.
        #. Explicitly closes this visual's figure window.
        #. Explicitly breaks all circular references between this figure and
           related artist objects (e.g., between this figure and its axes).
        #. Explicitly nullifies *all* attributes of the current object.
        #. Explicitly garbage collects.

        Caveats
        ----------
        This method should only be called:

        * If this visual is blocking (e.g., saved non-interactively or
          displayed interactively in a blocking manner). If this plot or
          animation is non-blocking, the resources assigned this plot or
          animation may be safely deallocated only *after* the end user closes
          the corresponding GUI widget or window.
        * As the last action of either:

          * The current caller.
          * The visual subclass.

        Attempting to subsequently call any other method *or* access any other
        variable bound to this object will reliably raise an exception.
        '''

        # If this figure still exists, attempt to safely close this figure.
        if self._figure is not None:
            mplfigure.close_figure(self._figure)

        # For each name and value of a field bound to this object...
        for field_name, field_value in objiter.iter_vars_custom_simple(self):
            # If this field itself contains a "figure" attribute, explicitly
            # nullify the latter to break this figure's circular references in
            # a manner ignoring "AttributeError: can't set attribute"
            # exceptions.
            #
            # Note that this probably fails to break all such references, as
            # doing so appears to be infeasible in a general-purpose manner.
            # These references are baked into the low-level Matplotlib API!
            try:
                if  hasattr(field_value, 'figure'):
                    setattr(field_value, 'figure', None)
            except AttributeError:
                pass

            # Explicitly nullify all attributes of this object, again ignoring
            # "AttributeError: can't set attribute" exceptions.
            try:
                setattr(self, field_name, None)
            except AttributeError:
                pass

        #FIXME: Unfortunately, explicitly garbage collecting immediately after
        #closing this animation appears to induce instabilities. Under the
        #"TkAgg" backend, for example, the following non-fatal runtime warning
        #is printed (presumably by the underlying Tcl/Tk library):
        #
        #    can't invoke "event" command:  application has been destroyed
        #        while executing
        #    "event generate $w <<ThemeChanged>>"
        #        (procedure "ttk::ThemeChanged" line 6)
        #        invoked from within
        #    "ttk::ThemeChanged"
        #
        #This suggests that if Python happens to garbage collect at this time,
        #the same issue will arise -- but hopefully much less frequently. Until
        #we resolve the underlying cause, we have no choice but to disable this.
        #Untrained glow bugs in a listless summer swelter!

        # Explicitly garbage collect.
        # gc.collect()

    # ..................{ PROPERTIES ~ read-only             }..................
    # Read-only properties, preventing callers from resetting these attributes.

    @property
    def phase(self) -> SimPhase:
        '''
        Current simulation phase.
        '''

        return self._phase


    @property
    def conf(self) -> SimConfVisualCellsABC:
        '''
        Configuration for this visual, synchronized with the user-defined
        YAML-backed simulation configuration file for this phase.
        '''

        return self._conf


    @property
    def kind(self) -> str:
        '''
        Machine-readable name identifying this type of visual (e.g.,
        ``voltage_extra``).
        '''

        return self._kind

    # ..................{ PROPERTIES ~ read-only : mpl       }..................
    @property
    def figure(self) -> Figure:
        '''
        Matplotlib figure for this visual, containing *all* matplotlib artists
        displayed by this visual (e.g., axes, colorbar).
        '''

        return self._axes


    @property
    def axes(self) -> Axes:
        '''
        Matplotlib axes for this visual's figure.

        All modelled variables for this cell cluster are spatially plotted onto
        this axes at each time step of this visual.
        '''

        return self._axes

    # ..................{ PROPERTIES ~ read-only : color     }..................
    #FIXME: Obsolete. Remove after adopting layers everywhere.
    @property
    def color_min(self) -> float | None:
        '''
        Minimum color value displayed on this visual's colorbar if this colorbar
        is *not* autoscaling colors *or* :data:`None` otherwise (i.e., if this
        colorbar is autoscaling colors).
        '''

        return self._color_min


    #FIXME: Obsolete. Remove after adopting layers everywhere.
    @property
    def color_max(self) -> float | None:
        '''
        Maximum color value displayed on this visual's colorbar if this colorbar
        is *not* autoscaling colors *or* :data:`None` otherwise (i.e., if this
        colorbar is autoscaling colors).
        '''

        return self._color_max


    #FIXME: Obsolete. Remove after adopting layers everywhere.
    @property
    def colormap(self) -> Colormap | None:
        '''
        Non-default matplotlib colormap if the external caller previously passed
        such a colormap via the optional ``colormap`` parameter to the
        :meth:`.__init__` method at visualization instantiation time *or*
        :data:`None` otherwise.

        This colormap maps all numeric data for one modelled variable for this
        cell cluster into color values displayed on this plot or animation's
        colorbar.
        '''

        return self._colormap

    # ..................{ MAKERS                             }..................
    @type_check
    def make_colorbar(self, color_mappable: ScalarMappable) -> None:
        '''
        Add a colorbar to this visual's figure associated with the passed
        matplotlib artist satisfying the color mappable API.

        This method is typically called by the **colorful layer** (i.e., layer
        displaying data intended to be coloured) defined for this visual.
        '''

        # If no colorbar title was passed to this visual, raise an exception.
        if not self._colorbar_title:
            raise BetseSimVisualException(
                'Animation "{}" parameter "colorbar_title" '
                'undefined or empty.'.format(self._kind))

        # Create and configure this colorbar.
        colorbar = self._figure.colorbar(color_mappable)
        colorbar.set_label(self._colorbar_title)

    # ..................{ LAYERS                             }..................
    @type_check
    def _append_layer(
        self,
        #FIXME: This class is dynamically unimportable by @beartype for some
        #reason -- probably a circular import dependency. *shrug*
        # *layers: 'betse.science.visual.layer.lyrabc.LayerCellsABC',
        *layers,
    ) -> None:
        '''
        Append all passed layers to the current sequence of layers, ensuring
        the subsequently called :meth:`_visualize_layers` method will visualize
        these layers after (and hence above) all previously appended layers.

        Parameters
        ----------
        layers : Tuple[LayerCellsABC]
            Tuple of all layers to be appended to this sequence of layers.
        '''

        self._layers.extend(layers)


    def _prep_layers(self) -> None:
        '''
        Prepare all layers to be subsequently layered onto this visual.

        If this simulation configuration requests that cells be labelled by
        their 0-based indices, a layer doing so is appended to the current
        layer sequence. Since this method is called *after* all user-defined
        layers have been appended, these labels are guaranteed to be layered
        over rather than under all plotted cell data.
        '''

        # If labelling each cell with its 0-based index, append a layer doing
        # so *AFTER* all lower layers (e.g., cell data) have been appended,
        if self._phase.p.visual.is_show_cell_indices:
            # Avoid circular import dependencies.
            from betse.science.visual.layer.lyrtext import LayerCellsIndex
            self._append_layer(LayerCellsIndex())

        # For the 0-based index of each layer and layer in the layer sequence
        # *AFTER* finalizing this sequence by appending final layers above...
        for layer_index, layer in enumerate(self._layers):
            # Z-order of all artists plotted by this layer. To prevent patch
            # artists defaulting to the default "ZORDER_PATCH" from being
            # plotted over artists plotted by the first layer, this z-order is
            # guaranteed to be at least as large as that default.
            layer_zorder = layer_index + ZORDER_PATCH

            # Prepare this layer.
            layer.prep(visual=self, zorder=layer_zorder)


    def _plot_layers(self) -> None:
        '''
        Iteratively plot all layers onto this visual for the
        current time step of this simulation.

        If this is:

        * The first time step, each such layer adds one or more Matplotlib
          artists to the figure for this visual.
        * Any time step _except_ the first, each such layer either:
          * Updates the contents of all artists previously added by that layer.
            This is the most efficient and hence ideal approach, but infeasible
            in cases in which updating artists is infeasible (e.g., the
            streamlines of a streamplot).
          * Removes all artists previously added by that layer _and_ adds new
            artists whose contents reflect this time step. This is the least
            efficient and hence least ideal approach.
        '''

        for layer in self._layers:
            layer.layer()

    # ..................{ COLORS                             }..................
    #FIXME: All methods in this subsection including this method are obsolete.
    #Remove after adopting layers everywhere. All other "FIXME:" comments below
    #are now ignorable, as this functionality should all go away.
    #FIXME: After doing so, also remove most color-specific attributes,
    #including:
    #
    #* "_color_mappables".
    #* "_color_min".
    #* "_color_max".
    #
    #The two notable exceptions are "_colormap" and "_colorbar_title", which
    #remain suitable for use in this base class.
    @type_check
    def _prep_colors_obsolete(
        self,

        # Obsolete optional parameters, which should no longer be passed.
        color_data: SequenceTypes,

        #FIXME: This is awful. For sanity, require callers always pass a
        #sequence of mappables: e.g.,
        #   color_mappables: IterableOrNoneTypes = None,
        #FIXME: Actually, this entire method is awful and should be refactored
        #away once layers are used everywhere. For backward compatilibility
        #with old-style animations, this is currently preserved as is.

        color_mappables: (ScalarMappable,) + IterableTypes,
    ) -> None:
        '''
        Prepare all color-specific data if this visual is colorful *or* noop
        otherwise (i.e., if this visual is colorless).

        If this visual is colorful, this method autoscale colors to the
        range implied by the passed color values *and* associate all mappables
        provided by colorful layers with this visuazilation's colorbar.

        Parameters
        ----------
        color_mappables : optional[ScalarMappable or IterableTypes]
            Iterable of all :class:`ScalarMappable` instances (e.g.,
            :class:`AxesImage`, :class:`ContourSet`) to associate with this
            visual's colorbar. For convenience, this parameter may
            be either:

            * A single mappable, in which case this colorbar will be associated
              with this mappable as is.
            * A non-string iterable (e.g., :class:`list`, :class:`set`) of one
              or more mappables, in which case this colorbar will be
              arbitrarily associated with the first mappable in this iterable.

            Defaults to ``None``, in which case this parameter defaults to the
            iterable of all mappables provided by the topmost and hence last
            mappable layer in the current layer sequence (i.e., the
            :meth:`LayerCellsColorfulABC.color_mappables` property of the
            last instance of the :class:`LayerCellsColorfulABC` subclass in
            the :attr:`_layers` attribute).
        color_data : optional[SequenceTypes]
            Multi-dimensional sequence of all color values to be plotted *or*
            ``None`` if calculating these values on initialization is
            impractical (e.g., due to space or time constraints). If colorbar
            autoscaling is enabled *and* this parameter is:

            * Non-``None``, the colorbar is clipped to the minimum and maximum
              scalar values unravelled from this array.
            * ``None``, the subclass is responsible for colorbar autoscaling.

            Defaults to `None`.
        '''

        # If a single mappable rather than a sequence of mappables was passed,
        # convert the former to the latter.
        if isinstance(color_mappables, ScalarMappable):
            color_mappables = (color_mappables,)

        # Autoscale colors to the range implied by the passed color values.
        self._autoscale_colors(color_data)

        # Associate these mappables with this visuazilation's colorbar *AFTER*
        # preparing all layers defining these mappables if no mappables were
        # passed.
        self._automap_colors(color_mappables)


    @type_check
    def _autoscale_colors(self, color_data: SequenceTypes) -> None:
        '''
        Autoscale the colorbar for this visual's figure to the
        minimum and maximum scalar values unravelled from the passed sequence
        by setting the :attr:`_color_min` and :attr:`_color_max` attributes to
        such values if colorbar autoscaling is both enabled and has not already
        been performed (i.e., these attributes are ``None``) *or* noop
        otherwise.

        Parameters
        ----------
        color_data : SequenceOrNoneTypes
            Multi-dimensional sequence of all color values to be plotted *or*
            ``None`` if calculating these values on initialization is
            impractical. See the :meth:`_prep_figure` method for further
            details.
        '''

        # If colorbar autoscaling is disabled, noop.
        if not self._conf.is_color_autoscaled:
            return

        # Flatten this multi-dimensional array to a one-dimensional array,
        # permitting efficient retrieval of minimum and maximum values.
        time_series_flat = np.ravel(color_data)

        # Set the current minimum and maximum color values.
        self._color_min = np.ma.min(time_series_flat)
        self._color_max = np.ma.max(time_series_flat)

        # Log these values.
        logs.log_debug(
            'Autoscaling animation colors to [%d, %d].',
            self._color_min, self._color_max)


    @type_check
    def _automap_colors(self, color_mappables: IterableTypes) -> None:
        '''
        Create a figure colorbar for this visual associated with the
        passed mappables if any or the mappables provided by the last mappable
        layer in the current layer sequence otherwise.

        Parameters
        ----------
        color_mappables : optional[IterableTypes]
            Iterable of all :class:`ScalarMappable` instances (e.g.,
            :class:`AxesImage`, :class:`ContourSet`) to associate with this
            visual's colorbar. By Matplotlib design, only the first
            mappable in this iterable is arbitrarily associated with this
            colorbar; all other mappables are ignored. Defaults to `None`, in
            which case the iterable of all mappables provided by the topmost
            and hence last mappable layer in the current layer sequence is
            defaulted to (i.e., the value of the
            :meth:`LayerCellsColorfulABC.color_mappables` property of the last
            instance of the :class:`LayerCellsColorfulABC` subclass in the
            :attr:`_layers` attribute).
        '''

        # Classify this sequence of mappables.
        self._color_mappables = color_mappables

        # Clip this mappable to the minimum and maximum colormap values *AFTER*
        # classifying this sequence.
        self._rescale_color_mappables()

        # First mappable in this iterable of mappables.
        color_mappable_first = iterget.get_item_first(self._color_mappables)

        # If this is *NOT* actually a mappable, raise an exception.
        objtest.die_unless_instance(
            obj=color_mappable_first, cls=ScalarMappable)

        # Create a colorbar associated with this mappable.
        self.make_colorbar(color_mappable_first)


    #FIXME: DRY: this method exactly recapitulates the existing
    #LayerCellsColorfulABC._scale_color_mappables() method. Consider excising
    #this method in favour of that more general-purpose method -- once we've
    #converted everything to leverage layers, anyway.
    def _rescale_color_mappables(self) -> None:
        '''
        Clip each color mappable previously passed to the
        :meth:`_automap_colors` method to the minimum and maximum colormap
        values.

        This method has the beneficial side-effect of establishing the range of
        this visual's colorbar.
        '''

        # For each previously passed color mappable...
        for color_mappable in self._color_mappables:
            # If this is *NOT* actually a mappable, raise an exception.
            objtest.die_unless_instance(
                obj=color_mappable, cls=ScalarMappable)

            # Clip this mappable to the minimum and maximum colormap values.
            color_mappable.set_clim(self._color_min, self._color_max)

    # ..................{ PLOTTERS                           }..................
    #FIXME: For generality, rename this method to visualize_time_step().
    #FIXME: For "PlotAfterSolving"-style plots, the first frame is uselessly
    #plotted twice. Investigate up this insanity, please. We might consider
    #simply ignoring the first call to this method if we can't get to the
    #ultimate bottom of this. Clearly, a matplotlib issue is implicated.

    @type_check
    def plot_frame(self, time_step: int) -> None:
        '''
        Display and/or save the frame corresponding to the passed sampled
        simulation time step for this visual.

        This method is iteratively called by matplotlib's
        :class:`FuncAnimation` class instantiated by our :meth:`_animate`
        method. The subclass implementation of this abstract method is expected
        to modify this animation's figure and axes so as to display the next
        frame. It is *not*, however, expected to save that figure to disk;
        frame saving is already implemented by this base class in a
        general-purpose manner.

        Specifically, this method (in order):

        #. Calls the subclass :meth:`_plot_frame_figure` method to update the
           current figure with this frame's data.
        #. Updates the current figure's axes title with the current time.
        #. Optionally writes this frame to disk if desired.

        Parameters
        ----------
        time_step : int
            0-based index of the frame to be plotted _or_ -1 if the most recent
            frame is to be plotted.
        '''

        # Absolute 0-based index of the frame to be plotted.
        #
        # If the passed index is -1 and hence relative rather than absolute,
        # this index is assumed to be the last index of the current
        # simulation's array of time steps.
        if time_step == -1:
            time_step_absolute = len(self._phase.sim.time) - 1
        # Else, the passed index is already absolute and hence used as is.
        else:
            time_step_absolute = time_step

        # Log this animation frame.
        logs.log_debug(
            'Exporting "%s" frame %d / %d...',
            self._kind, time_step_absolute, self._time_step_last)

        # Classify this time step for subsequent access by subclasses.
        self._time_step = time_step

        # Plot this frame's title *BEFORE* this frame, allowing axes changes
        # performed by the subclass implementation of the _plot_frame_figure()
        # method called below to override the default title.
        self._plot_frame_axes_title()

        # Plot this frame via external layers *BEFORE* plotting this frame
        # via subclass logic, allowing the latter to override the former.
        self._plot_layers()

        # Plot this frame via subclass logic *AFTER* performing all
        # superclass-specific plotting.
        self._plot_frame_figure()

        # Interactively display this frame if requested.
        self._show_frame(time_step_absolute)

        # Non-interactively write this frame to disk if requested.
        self._save_frame(time_step_absolute)

        # Inform subclasses that subsequent frames are no longer the first.
        self._is_time_step_first = False


    def _plot_frame_axes_title(self) -> None:
        '''
        Plot the current frame title of this animation onto this animation's
        axes.

        By default, this title interpolates the current time step and must thus
        be replotted for each animation frame.
        '''

        #FIXME: Shift into a new "betse.util.time.times" submodule.

        # Number of seconds in a minute.
        SECONDS_PER_MINUTE = 60

        # Number of seconds in an hour.
        SECONDS_PER_HOUR = SECONDS_PER_MINUTE * 60

        #FIXME: For efficiency, classify the following local variables as
        #object attributes in the __init__() method rather than recompute these
        #variables each animation frame here.

        # Human-readable suffix of the units that simulation times are reported
        # in (e.g., "ms" for milliseconds).
        time_unit_suffix = None

        # Factor by which low-level simulation times are multiplied to yield
        # human-readable simulation times in the same units of
        # "time_unit_suffix".
        time_unit_factor = None

        # Duration in seconds of the current simulation phase.
        time_len = self._phase.p.total_time

        # If this phase runs for less than or equal to 100ms, report
        # simulation time in milliseconds (i.e., units of 0.001s).
        if time_len <= 0.1:
            time_unit_suffix = 'ms'
            time_unit_factor = 1e3
        # Else if this phase runs for less than or equal to one minute, report
        # simulation time in seconds (i.e., units of 1s).
        elif time_len <= SECONDS_PER_MINUTE:
            time_unit_suffix = 's'
            time_unit_factor = 1
        # Else if this phase runs for less than or equal to one hour, report
        # simulation time in minutes (i.e., units of 60s).
        elif time_len <= SECONDS_PER_HOUR:
            time_unit_suffix = ' minutes'
            time_unit_factor = 1/SECONDS_PER_MINUTE
        # Else, this phase is assumed to run for less than or equal to one day.
        # In this case, simulation time is reported in hours (i.e., units of
        # 60*60s).
        else:
            time_unit_suffix = ' hours'
            time_unit_factor = 1/SECONDS_PER_HOUR

        # Current time adjusted for long/short simulation.
        time_accelerated = (
            time_unit_factor *
            self._phase.sim.time[self._time_step])

        # Update this figure with this time, rounded to one decimal place.
        self._axes.set_title('{} (time: {:.1f}{})'.format(
            self._axes_title, time_accelerated, time_unit_suffix,))


    def _show_frame(self, time_step_absolute: int) -> None:
        '''
        Display the current frame of this animation if requested by the current
        simulation configuration *or* silently noop otherwise.

        Parameters
        ----------
        time_step_absolute : int
            0-based index of the current frame to be shown.
        '''

        # If *NOT* showing this frame, return immediately.
        if not self._is_show:
            return

        # Temporarily yield the time slice for the smallest amount of time
        # required by the current matplotlib backend to handle queued events in
        # the GUI-specific event loop of the current process. Failing to do so
        # produces subtle issues with plot and animation display. Technically,
        # doing so only appears to be required on one or more of the following
        # edge-case conditions:
        #
        # * The current platform is Windows, whose POSIX-incompatible process
        #   model has special needs.
        # * The current visual is non-blocking *AND* the current
        #   backend is poorly implemented or supported (e.g., "Qt4Agg").
        #
        # Since reliably detecting all such conditions is non-trivial and
        # certainly more time consuming than unconditionally yielding the time
        # slice for a negligible amount of time, we prefer the latter approach.
        #
        # In theory, the time slice should *NEVER* need to be manually yielded.
        # The "matplotlib.animation.FuncAnimation" class should handle such
        # low-level details, as confirmed by the pyplot.pause() docstring:
        #
        #  pause(interval)
        #
        #     Pause for *interval* seconds.
        #
        #     If there is an active figure it will be updated and displayed,
        #     and the gui event loop will run during the pause.
        #
        #     If there is no active figure, or if a non-interactive backend
        #     is in use, this executes time.sleep(interval).
        #
        #     This can be used for crude animation. For more complex
        #     animation, see :mod:`matplotlib.animation`.
        #
        # Note the last paragraph, implying the low-level pyplot.pause() method
        # to be an alternative (rather than addition) to the the higher-level
        # "matplotlib.animation" submodule. Unfortunately, this does *NOT*
        # appear to be the case; the pyplot.pause() method *MUST* be manually
        # called even when leveraging "matplotlib.animation" functionality.
        #
        # For further details, see also:
        #
        #     https://gitlab.com/betse/betse/issues/9
        #     https://github.com/matplotlib/matplotlib/issues/2134/
        #
        # When yielding the time slice, the following non-fatal deprecation
        # warning emitted by the pyplot.pause() function is temporarily
        # suppressed:
        #
        #     [betse] /usr/lib64/python3.4/site-packages/matplotlib/backend_bases.py:2437:
        #     MatplotlibDeprecationWarning: Using default event loop
        #     until function specific to this GUI is implemented
        #       warnings.warn(str, mplDeprecation)
        #
        # Since this warning is overly verbose and safely ignorable, we do so.
        # While mildly inefficient, ignoring this warning is preferable to the
        # alternatives of enraging, berating, and frightening end users.
        with ignoring_deprecations_mpl():
            pyplot.pause(0.0001)


    def _save_frame(self, time_step_absolute: int) -> None:
        '''
        Persist the current frame of this animation to disk if requested by the
        current simulation configuration *or* silently noop otherwise.

        Parameters
        ----------
        time_step_absolute : int
            0-based index of the current frame to be saved.
        '''

        # If *NOT* saving this frame, return immediately.
        if not self._is_save:
            return

        # If saving animation frames as images, save this frame as such.
        if self._writer_images is not None:
            self._writer_images.grab_frame(**self._writer_savefig_kwargs)

        # If saving animation frames as video, save this frame as such.
        if self._writer_video is not None:
            # For debuggability, temporarily escalate the
            # matplotlib-specific verbosity level.
            with mpl_config.reducing_log_level_to_debug_if_info():
                self._writer_video.grab_frame(
                    **self._writer_savefig_kwargs)

        # If this is the last frame to be plotted, finalize all writers
        # *AFTER* instructing these writers to write this frame.
        if time_step_absolute == self._time_step_last:
            logs.log_debug(
                'Finalizing animation "%s" saving...', self._kind)
            self._close_writers()

    # ..................{ SUBCLASS                           }..................
    def _plot_frame_figure(self) -> None:
        '''
        Update this visual's figure (and typically axes) content to reflect the
        current simulation time step.

        This method defaults to a noop. Subclasses may optionally redefine this
        method with subclass-specific logic but are strongly encouraged to
        append layers containing such logic instead.
        '''

        pass

    # ..................{ PLOTTERS ~ image                   }..................
    #FIXME: Merge the useful docstring for this method into the
    #"betse.science.visual.layer.field.layerfieldsurface.LayerCellsFieldSurface"
    #class; then, excise this method.
    @type_check
    def _plot_image(
        self,

        # Mandatory parameters.
        pixel_data: SequenceTypes,

        # Optional parameters.
        colormap : Colormap = None,
    ) -> AxesImage:
        '''
        Plot and return an image of the passed pixel data onto the current
        figure's axes.

        Parameters
        ----------
        pixel_data: np.ndarray
            Array of pixel data defining the image to be plotted, whose first
            two dimensions index the X and Y components of this axes. If this
            array is:

            * Two-dimensional, a greyscale image mapped onto the passed
              colormap will be plotted.
            * Three-dimensional, an RGB image *not* mapped onto the passed
              colormap will be plotted.
            * Four-dimensional, an RGBa image *not* mapped onto the passed
              colormap will be plotted.

        colormap : matplotlib.cm.Colormap
            Optional colormap with which to map the passed pixel data when
            greyscale (and ignored otherwise) *or* ``None``, in which case the
            default colormap will be used.

        Returns
        ----------
        matplotlib.image.AxesImage
            Image produced by plotting the passed pixel data.

        See Also
        ----------
        http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.imshow
            Further details on Matplotlib-based axes image plotting.
        '''

        # Default unpassed parameters.
        if colormap is None:
            colormap = self._colormap

        # Plot and return this image.
        return self._axes.imshow(
            pixel_data,
            origin='lower',
            extent=self._phase.cache.upscaled.extent,
            cmap=colormap,
        )


    #FIXME: Replace entirely by the appropriate "LayerCellsFieldStream" subclass.
    @type_check
    def _plot_stream(
        self,

        # Mandatory arguments.
        x: SequenceTypes,
        y: SequenceTypes,
        magnitude: SequenceTypes,

        # Optional arguments.
        grid_x: SequenceOrNoneTypes = None,
        grid_y: SequenceOrNoneTypes = None,
        magnitude_max: NumericOrNoneTypes = None,
        old_stream_plot: (StreamplotSet, NoneType) = None,
    ) -> StreamplotSet:
        '''
        Plot and return a streamplot of all streamlines of the passed vector
        flow data onto the current figure's axes.

        Parameters
        ----------
        x : SequenceTypes
            Two-dimensional sequence of X components of vector flow velocity.
        y : SequenceTypes
            Two-dimensional sequence of Y components of vector flow velocity.
        magnitude: SequenceTypes
            One-dimensional sequence of vector flow magnitudes.
        grid_x : SequenceTypes or NoneType
            Optional scaled X components of the cell cluster grid. Defaults to
            `None`, in which case the default `self.cells.X` array is used.
        grid_y : SequenceTypes or NoneType
            Optional scaled Y components of the cell cluster grid. Defaults to
            `None`, in which case the default `self.cells.Y` array is used.
        magnitude_max: NumericSimpleTypes or NoneType
            Optional maximum magnitude in the passed `magnitude` array.
            Defaults to `None`, in which case this array is implicitly searched
            for this value.
        old_stream_plot: StreamplotSet or NoneType
            Optional streamplot returned by a prior call to this method
            (typically for a prior frame) _or_ `None` if this is the first call
            to this method for this animation. If non-`None`, this streamplot
            will be cleared in preparation for re-streamplotting by this call.

        Returns
        ----------
        StreamplotSet
            Streamplot produced by plotting the passed vector flow data.

        See Also
        ----------
        http://matplotlib.org/api/axes_api.html#matplotlib.axes.Axes.streamplot
            Further details on Matplotlib-based axes streamplotting.
        '''

        # Default all unpassed optional arguments.
        if magnitude_max is None:
            magnitude_max = np.max(magnitude)
        if grid_x is None:
            grid_x = self._phase.cells.X * self._phase.p.um
        if grid_y is None:
            grid_y = self._phase.cells.Y * self._phase.p.um

        # If a prior streamplot to be erased was passed, do so.
        if old_stream_plot is not None:
            # Erase this streamplot's streamlines before replotting.
            old_stream_plot.lines.remove()

            # If this version of Matplotlib supports erasing the set of patches
            # corresponding to this streamplot's arrow heads, do so.
            try:
                old_stream_plot.arrows.remove()
            # Else, manually erase them by iterating over all patch objects and
            # preserving all non-arrow head patches. This will also remove all
            # arrow head patches of streamplots plotted by this subclass for
            # this frame, which is non-ideal. Matplotlib leaves us no choice.
            except NotImplementedError:
                self._axes.patches = [
                    patch
                    for patch in self._axes.patches
                    if not isinstance(patch, FancyArrowPatch)
                ]

        # Plot and return this streamplot.
        return self._axes.streamplot(
            grid_x, grid_y, x, y,
            density=self._phase.p.stream_density,
            linewidth=(3.0*magnitude/magnitude_max) + 0.5,
            color=self._phase.p.vcolor,
            cmap=self._colormap,
            # arrowsize=3.0,

            # Draw this streamplot over all patch and line artists, by default.
            # See the "ZORDER_STREAM" docstring for further commentary.
            zorder=ZORDER_STREAM,
        )

    # ..................{ PLOTTERS ~ cell                    }..................
    #FIXME: Pretty intense, and obviously better refactored two distinct
    #"LayerCellsABC" subclasses. This will probably prove pivotal to
    #implementing deformations sanely.
    #FIXME: After doing so, excise *ALL* of the methods below.

    def _plot_cells_sans_ecm(self, *args, **kwargs) -> PolyCollection:
        '''
        Plot and return an intracellular plot of all cells with colours
        corresponding to the passed vector of arbitrary cell data (e.g.,
        transmembrane cell voltages for the current time step) onto the current
        figure's axes.

        The type of plot returned is defined by this simulation's
        configuration. Specifically:

        * If this configuration requests that individual cells be plotted
          (i.e., ``show cells`` is ``True``), a mosaic plot of this
          simulation's cell cluster will be returned. See the
          :meth:`_plot_cell_mosaic` method.
        * Else, a mesh plot interpolating the individual cells of this
          simulation's cell cluster will be returned. See the
          `_plot_cell_mesh()` method.

        Parameters
        -----------
        cell_data : np.ndarray
            Arbitrary cell data defined on an environmental grid to be plotted.

        All other passed parameters will be passed as is to the underlying plot
        method called by this method (e.g., `_plot_cell_mosaic()`).

        Returns
        -------
        PolyCollection
            Plot produced by plotting the passed cell data.
        '''

        if self._phase.p.showCells is True:
            return self._plot_cell_mosaic(*args, **kwargs)
        else:
            return self._plot_cell_mesh(*args, **kwargs)


    def _update_cell_plot_sans_ecm(
        self,
        cell_plot: PolyCollection,
        cell_data: np.ndarray,
        *args, **kwargs
    ) -> None:
        '''
        Update _without_ recreating the passed intracellular cell plot with the
        passed vector of arbitrary cell data (e.g., transmembrane cell voltages
        for the current time step) onto the current figure's axes.

        Parameters
        -----------
        cell_plot : PolyCollection
            Cell plot previously returned by either the
            :meth:`_plot_cells_sans_ecm` *or*
            :meth:`_revive_cell_plot_sans_ecm` method.
        cell_data : np.ndarray
            Arbitrary cell data defined on an environmental grid to be plotted.

        All other passed parameters will be passed as is to the underlying plot
        method called by this method (e.g., :meth:`_plot_cell_mosaic`).
        '''
        assert types.is_sequence_nonstr(cell_data), (
            types.assert_not_sequence_nonstr(cell_data))

        # If plotting individuals cells, update this plot with this data as is.
        if self._phase.p.showCells:
            assert types.is_matplotlib_polycollection(cell_plot), (
                types.assert_not_matplotlib_polycollection(cell_plot))
            cell_plot.set_array(cell_data)
        # Else, the cell cluster is being plotted as an unstructured triangular
        # grid, requiring this data be reshaped onto this grid.
        else:
            assert types.is_matplotlib_trimesh(cell_plot), (
                types.assert_not_matplotlib_trimesh(cell_plot))

            # membranes_midpoint_data = np.zeros(len(self.cells.voronoi_centres))
            # membranes_midpoint_data[self.cells.cell_to_grid] = membranes_midpoint_data

            # Update this plot with this gridded data.
            cell_plot.set_array(cell_data)


    def _revive_cell_plots_sans_ecm(
        self,
        cell_plot: PolyCollection,
        cell_data: np.ndarray,
        *args, **kwargs
    ) -> PolyCollection:
        '''
        Recreate the passed intracellular plot and return a similar plot of all
        cells with colours corresponding to the passed vector of arbitrary cell
        data (e.g., transmembrane cell voltages for the current time step) onto
        the current figure's axes.

        This method is typically called to replot an animation of a cell cluster
        subject to physical cell changes (e.g., cutting, deformation). The type
        of plot returned is defined by this simulation's configuration.
        Specifically:

        * If this configuration requests that individual cells be plotted (i.e.,
          ``show cells`` is ``True``), the passed plot _must_ be a mosaic plot
          previously returned by the :meth:`_plot_cell_mesh` method. This plot
          will be updated in-place and returned as is without creating a new
          plot.
        * Else, the passed plot _must_ be a mesh plot previously returned by the
          :meth:`_plot_cell_mesh` method -- specifically, a :class:`TriMesh`
          object. Since the :class:`TriMesh` API currently provides no public
          means of updating such plots in-place, this method (in order):

          #. Removes the passed mesh plot from this figure's axes.
          #. Creates a new mesh plot and adds that plot to this figure's axes.
          #. Returns that plot.

        Parameters
        -----------
        cell_plot : PolyCollection
            Cell plot previously returned by either the
            :meth:`_plot_cells_sans_ecm` *or*
            :meth:`_revive_cell_plots_sans_ecm` method.
        cell_data : np.ndarray
            Arbitrary cell data defined on an environmental grid to be plotted.

        All other passed parameters will be passed as is to the underlying plot
        method called by this method (e.g., :meth:`_plot_cell_mosaic`).

        Returns
        --------
        PolyCollection
            Plot produced by replotting the passed cell data.
        '''
        assert types.is_sequence_nonstr(cell_data), (
            types.assert_not_sequence_nonstr(cell_data))

        # If plotting individual cells, the passed cell plot *MUST* be a
        # polygon collection previously returned by the _plot_cell_mosaic()
        # method.
        if self._phase.p.showCells:
            assert types.is_matplotlib_polycollection(cell_plot), (
                types.assert_not_matplotlib_polycollection(cell_plot))

            # Update this plot in-place.
            cell_plot.set_array(cell_data)
            cell_plot.set_verts(
                nparray.from_iterable(self._phase.cells.cell_verts) *
                self._phase.p.um)

            # Return the same plot.
            return cell_plot
        # Else, the passed cell plot *MUST* be a triangle mesh previously
        # returned by the _plot_cell_mesh() method.
        else:
            assert types.is_matplotlib_trimesh(cell_plot), (
                types.assert_not_matplotlib_trimesh(cell_plot))

            cell_plot.remove()
            return self._plot_cell_mesh(cell_data=cell_data, *args, **kwargs)


    def _plot_cell_mosaic(self, cell_data: SequenceTypes) -> PolyCollection:
        '''
        Plot and return a mosaic plot of all cells with colours corresponding
        to the passed vector of arbitrary cell data (e.g., transmembrane
        voltages for all cells for the current time step) onto the current
        figure's axes.

        The returned plot will be a polygon collection such that each polygon
        signifies a cell in this simulation's cell cluster.

        Parameters
        -----------
        cell_data : SequenceTypes
            Arbitrary cell data defined on an environmental grid to be plotted.

        Returns
        --------
        PolyCollection
            Mosaic plot produced by plotting the passed cell data.
        '''

        # Cell vertices plotted as polygons.
        mosaic_plot = PolyCollection(
            verts=mathunit.upscale_coordinates(self._phase.cells.cell_verts),
            cmap=self._colormap,
            edgecolors='none',
        )

        # Associate this plot with the passed cell data. Curiously, the
        # Matplotlib API provides no means of passing this data to the
        # PolyCollection.__init__() constructor. (Yes, we checked. Twice.)
        mosaic_plot.set_array(cell_data)

        # Add this plot to this figure's axes.
        self._axes.add_collection(mosaic_plot)
        return mosaic_plot


    #FIXME: This plots somewhat similarly to the presumably superior
    #"LayerCellsVectorSurfaceDiscrete" subclass. Generalize this method into a
    #new "LayerCellsGouraudContinuous" subclass of the same submodule.
    # FIXME: I like this version better than the current one that animates
    # things like Vmem when p.showCells is False
    def _plot_cell_mesh(self, cell_data: np.ndarray) -> 'TriMesh':
        '''
        Plot and return a mesh plot of all cells with colours corresponding to
        the passed vector of arbitrary cell data (e.g., transmembrane voltages
        for all cells for the current time step) onto the current figure's
        axes.

        The returned plot will be an unstructured triangular grid interpolating
        each cell of this simulation's cell cluster into a smooth continuum.

        Parameters
        -----------
        cell_data : np.ndarray
            Arbitrary cell data defined on an environmental grid to be plotted.

        Returns
        --------
        TriMesh
            Mesh plot produced by plotting the passed cell data.
        '''
        assert types.is_sequence_nonstr(cell_data), (
            types.assert_not_sequence_nonstr(cell_data))

        # If the passed cell data is defined on membrane midpoints, average
        # that to correspond to cell centres instead.
        if len(cell_data) == len(self._phase.cells.mem_i):
            # cell_data = np.dot(
            #     self._phase.cells.M_sum_mems, cell_data) / self._phase.cells.num_mems
            xi = self._phase.p.um*self._phase.cells.mem_mids_flat[:,0]
            yi = self._phase.p.um*self._phase.cells.mem_mids_flat[:, 1]

        else:

            xi = self._phase.p.um*self._phase.cells.cell_centres[:,0]
            yi = self._phase.p.um*self._phase.cells.cell_centres[:, 1]

        # Unstructured triangular grid assigned the passed cell data.
        # triangular_grid = np.zeros(len(self._phase.cells.voronoi_centres))
        # triangular_grid[self._phase.cells.cell_to_grid] = cell_data

        # cmin = membranes_midpoint_data.min()
        # cmax = membranes_midpoint_data.max()

        # Create and add this plot to this figure's axes and return this plot.
        # Unfortunately, the tripcolor() function requires:
        #
        # * The "x", "y", and "triangles" parameters to be positional.
        # * All other parameters to be keyword.
        #
        # Behold! The ultimate examplar of nonsensical API design.
        return self._axes.tripcolor(
            xi,
            yi,
            cell_data,
            shading='gouraud',
            cmap=self._colormap
        )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **non-string iterable getters** (i.e., utility functions getting one
or more items contained within non-string objects implementing the abstract
:class:`collections.abc.Iterable` base class) facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseIterableException, BetseParamException
from betse.util.io.log import logs
from betse.util.type.types import (
    type_check,
    CallableTypes,
    IterableTypes,
    StrOrNoneTypes,
    TestableTypes,
)

# ....................{ GETTERS                           }....................
@type_check
def get_items_duplicate(iterable: IterableTypes) -> set:
    '''
    Unordered set of all duplicate items of the passed iterable.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.

    Returns
    ----------
    set
        Unordered set of all duplicate items of this iterable.

    See Also
    ----------
    https://stackoverflow.com/a/9835819/2809027
        Stackoverflow answer strongly inspiring this implementation.
    '''

    # Set of all unique items of this iterable previously visited below.
    items_unique = set()

    # Set of all duplicate items of this iterable to be returned.
    items_duplicate = set()

    # For each item of this iterable...
    for item in iterable:
        # If this item has been previously visited, this item is a duplicate.
        if item in items_unique:
            items_duplicate.add(item)
        # Else, this item is unique.
        else:
            items_unique.add(item)

    # Return the set of all duplicate items.
    return items_duplicate

# ....................{ GETTERS ~ first                   }....................
@type_check
def get_item_first(*iterables: IterableTypes) -> object:
    '''
    First item non-destructively retrieved from the first passed non-empty
    iterable if at least one such iterable is non-empty *or* raise an exception
    otherwise (i.e., if all passed iterables are empty).

    Parameters
    ----------
    iterables : tuple[IterableTypes]
        Tuple of all iterables to be inspected.

    Returns
    ----------
    object
        First item of the first passed non-empty iterable.

    Raises
    ----------
    BetseIterableException
        If all passed iterables are empty.

    See Also
    ----------
    :func:`get_item_first_or_sentinel`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # First item of these iterables if any *OR* the sentinel placeholder,
    item_first = get_item_first_or_sentinel(*iterables)

    # If no such item exists, raise an exception.
    if item_first is SENTINEL:
        raise BetseIterableException('Iterables empty.')
    # Else, this item exists.

    # Return this item.
    return item_first


@type_check
def get_item_first_or_sentinel(*iterables: IterableTypes) -> object:
    '''
    First item non-destructively retrieved from the first passed non-empty
    iterable if at least one such iterable is non-empty *or* the sentinel
    placeholder otherwise (i.e., if all passed iterables are empty).

    Specifically, if the first passed non-empty iterable is a:

    * Sequence (e.g., :class:`list`, :class:`tuple`), this function returns the
      first item of this sequence.
    * Non-sequence (e.g., :class:`set`, :class:`dict`), this function returns a
      pseudo-random item of this non-sequence. While most non-sequences
      guarantee predictable order of retrieval assuming no intervening changes,
      this is a fairly unreliable assumption.

    Parameters
    ----------
    iterables : tuple[IterableTypes]
        Tuple of all iterables to be inspected.

    Returns
    ----------
    object
        Either:

        * If at least one such iterable is non-empty, the first item of this
          iterable.
        * Else, the sentinel singleton.

    See Also
    ----------
    https://stackoverflow.com/a/40054478/2809027
        Stackoverflow answer strongly inspiring this implementation, complete
        with detailed timings of all alternative solutions.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL
    from betse.util.type.iterable import iterables as itermod

    # First item of these iterables if any *OR* the sentinel placeholder,
    # defaulting to the latter for implementation simplicity.
    item_first = SENTINEL

    # If at least one such iterable is non-empty, iterate this iterable once
    # and assign the first item of this iterable to this local when doing so.
    #
    # Note that this implementation, despite being verbose and unwieldy, has
    # been meticulously timed to be the most space and time efficient... Yup!
    for item_first in itermod.iter_items(*iterables):
        break

    # Return this item.
    return item_first

# ....................{ GETTERS ~ first : instance        }....................
@type_check
def get_item_first_instance_of(
    iterable: IterableTypes, cls: TestableTypes, **kwargs) -> object:
    '''
    First instance of the passed class or tuple of classes retrieved from the
    passed iterable if this iterable contains such an item *or* raise an
    exception otherwise (i.e., if this iterable contains no such item).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be searched.
    cls : TestableTypes
        Class or tuple of classes of the item to search this iterable for.
    kwargs : dict
        Dictionary of all remaining keyword arguments to be passed as is to the
        :func:`get_item_first_satisfying` function.

    Returns
    ----------
    object
        First instance of this class in this iterable.

    Raises
    ----------
    BetseIterableException
        If this iterable contains no such item.

    See Also
    ----------
    :func:`get_item_first_satisfying_or_sentinel`
        Further details on ordering guarantees.
    '''

    return get_item_first_satisfying(
        iterable=iterable,
        predicate=lambda item: isinstance(item, cls),
        **kwargs
    )


@type_check
def get_item_first_not_instance_of(
    iterable: IterableTypes, cls: TestableTypes, **kwargs) -> object:
    '''
    First item of this iterable that is *not* an instance of the passed class
    or tuple of classes if this iterable contains such an item *or* raise an
    exception otherwise (i.e., if all items of this iterable are such
    instances).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be searched.
    cls : TestableTypes
        Class or tuple of classes of the item to search this iterable for.
    kwargs : dict
        Dictionary of all remaining keyword arguments to be passed as is to the
        :func:`get_item_first_satisfying` function.

    Returns
    ----------
    object
        First non-instance of this class in this iterable.

    Raises
    ----------
    BetseIterableException
        If this iterable contains only such instances.

    See Also
    ----------
    :func:`get_item_first_satisfying_or_sentinel`
        Further details on ordering guarantees.
    '''

    return get_item_first_satisfying(
        iterable=iterable,
        predicate=lambda item: not isinstance(item, cls),
        **kwargs
    )

# ....................{ GETTERS ~ first : satisfying      }....................
@type_check
def get_item_first_satisfying(
    iterable: IterableTypes,
    predicate: CallableTypes,
    exception_message: str = None,
) -> object:
    '''
    First item of the passed iterable satisfying the passed **predicate**
    (i.e., callable accepting one parameter, returning ``True`` only if this
    parameter suffices) if this iterable contains such an item *or* raise an
    exception otherwise (i.e., if this iterable contains no such item).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.
    predicate : CallableTypes
        Callable accepting one parameter and returning ``True`` only if this
        parameter suffices.
    exception_message : optional[str]
        Exception message to be raised if no such item is found. Defaults to
        ``None``, in which case a general-purpose message is synthesized.

    Returns
    ----------
    object
        First item satisfying this predicate in this iterable.

    Raises
    ----------
    BetseIterableException
        If this iterable contains no such item.

    See Also
    ----------
    :func:`get_item_first_satisfying_or_sentinel`
        Further details on ordering guarantees.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # First item satisfying this predicate in this iterable if any *OR* the
    # sentinel placeholder otherwise.
    first_item = get_item_first_satisfying_or_sentinel(iterable, predicate)

    # If no item satifies this predicate, raise an exception.
    if first_item is SENTINEL:
        # If no exception message is passed, synthesize a default message.
        if exception_message is None:
            exception_message = (
                'Iterable "{}" item satisfying predicate {} not found.'.format(
                    iterable, predicate))

        # Raise this exception.
        raise BetseIterableException(exception_message)

    # Else, return this element.
    return first_item


@type_check
def get_item_first_satisfying_or_sentinel(
    iterable: IterableTypes, predicate: CallableTypes) -> object:
    '''
    First item of the passed iterable satisfying the passed **predicate**
    (i.e., callable accepting one parameter, returning ``True`` only if this
    parameter suffices) if this iterable contains such an item *or* the
    **sentinel singleton** (i.e.,
    :attr:`betse.util.type.obj.sentinels.SENTINEL`) otherwise.

    If the passed iterable is a:

    * Sequence, this is guaranteed to be the first such element.
    * Non-sequence (e.g., :class:`set`, :class:`dict`), this should be assumed
      to be a random such element. While most non-sequences guarantee
      predictable order of retrieval assuming no intervening changes, this is a
      fairly unreliable assumption.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.
    predicate : CallableTypes
        Callable accepting one parameter and returning ``True`` only if this
        parameter suffices.

    Returns
    ----------
    object
        Either:

        * If one or more items of this iterable satisfy this predicate, the
          first such item.
        * Else, the sentinel singleton.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Collective efficiency is our middle names.
    return next((item for item in iterable if predicate(item)), SENTINEL)

# ....................{ GETTERS ~ last : instance         }....................
@type_check
def get_item_last_instance_of(
    iterable: IterableTypes, cls: TestableTypes, **kwargs) -> object:
    '''
    Last instance of the passed class or tuple of classes retrieved from the
    passed iterable if this iterable contains such an item *or* raise an
    exception otherwise (i.e., if this iterable contains no such item).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be searched.
    cls : TestableTypes
        Class or tuple of classes of the item to search this iterable for.
    kwargs : dict
        Dictionary of all remaining keyword arguments to be passed as is to the
        :func:`get_item_last_satisfying` function.

    Returns
    ----------
    object
        Last instance of this class in this iterable.

    Raises
    ----------
    BetseIterableException
        If this iterable contains no such item.

    See Also
    ----------
    :func:`get_item_last_satisfying`
        Further details on ordering guarantees.
    '''

    return get_item_last_satisfying(
        iterable=iterable,
        predicate=lambda item: isinstance(item, cls),
        **kwargs
    )


@type_check
def get_item_last_instance_of_or_none(
    iterable: IterableTypes, cls: TestableTypes, **kwargs) -> object:
    '''
    Last instance of the passed class or tuple of classes retrieved from the
    passed iterable if this iterable contains such an item *or* ``None``
    otherwise (i.e., if this iterable contains no such element).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be searched.
    cls : TestableTypes
        Class or tuple of classes of the item to search this iterable for.
    kwargs : dict
        Dictionary of all remaining keyword arguments to be passed as is to the
        :func:`get_item_last_satisfying` function.

    Returns
    ----------
    object
        Last instance of this class in this iterable if any *or*
        :attr:`betse.util.type.obj.sentinels.SENTINEL` otherwise.

    Raises
    ----------
    BetseIterableException
        If this iterable contains no such item.

    See Also
    ----------
    :func:`get_item_last_satisfying`
        Further details on ordering guarantees.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Last instance of this class in this iterable if any or the sentinel
    # placeholder otherwise.
    item_found = get_item_last_satisfying_or_sentinel(
        iterable=iterable,
        predicate=lambda item: isinstance(item, cls),
        **kwargs
    )

    # Return this instance if *NOT* the sentinal placeholder or None otherwise.
    return item_found if item_found is not SENTINEL else None

# ....................{ GETTERS ~ last : satisfying       }....................
@type_check
def get_item_last_satisfying(
    iterable: IterableTypes,
    predicate: CallableTypes,
    exception_message: str = None,
) -> object:
    '''
    Last element of the passed iterable satisfying the passed **predicate**
    (i.e., callable accepting one parameter, returning ``True`` only if this
    parameter suffices) if this iterable contains such an element *or* raise an
    exception otherwise (i.e., if this iterable contains no such element).

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.
    predicate : CallableTypes
        Callable accepting one parameter and returning ``True`` only if this
        parameter suffices.
    exception_message : optional[str]
        Exception message to be raised if no such element is found. Defaults to
        ``None``, in which case a suitably general-purpose message is
        synthesized.

    Returns
    ----------
    object
        Last element satisfying this predicate in this iterable.

    Raises
    ----------
    BetseIterableException
        If this iterable contains no such element.

    See Also
    ----------
    :func:`get_item_first_satisfying_or_sentinel`
        Further details on ordering guarantees.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable import iterables

    # For simplicity, the existing get_item_first_satisfying() function is
    # deferred to by returning the first element in the reverse of this
    # iterable satisfying this predicate.
    return get_item_first_satisfying(
        # For safety, this iterable is reversed via the high-level reverse()
        # function rather than the low-level reversed() builtin; the latter
        # fails to generically support all possible iterable types.
        iterable=iterables.reverse(iterable),
        predicate=predicate,
        exception_message=exception_message,
    )


@type_check
def get_item_last_satisfying_or_sentinel(
    iterable: IterableTypes, predicate: CallableTypes) -> object:
    '''
    Last element of the passed iterable satisfying the passed **predicate**
    (i.e., callable accepting one parameter, returning ``True`` only if this
    parameter suffices) if this iterable contains such an element *or* the
    :attr:`betse.util.type.obj.sentinels.SENTINEL` placeholder constant
    otherwise.

    If the passed iterable is a:

    * Sequence, this is guaranteed to be the last such element.
    * Non-sequence (e.g., :class:`set`, :class:`dict`), this should be assumed
      to be a random such element. While most non-sequences guarantee
      predictable order of retrieval assuming no intervening changes, this is a
      fairly unreliable assumption.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.
    predicate : CallableTypes
        Callable accepting one parameter and returning ``True`` only if this
        parameter suffices.

    Returns
    ----------
    object
        Last element satisfying this predicate in this iterable if any *or*
        :attr:`betse.util.type.obj.sentinels.SENTINEL` otherwise.

    Raises
    ----------
    BetseIterableException
        If this iterable contains no such element.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable import iterables

    # For simplicity, the existing get_item_first_satisfying_or_sentinel()
    # function is deferred to by returning the first element in the reverse of
    # this iterable satisfying this predicate.
    return get_item_first_satisfying_or_sentinel(
        # For safety, this iterable is reversed via the high-level reverse()
        # function rather than the low-level reversed() builtin; the latter
        # fails to generically support all possible iterable types.
        iterable=iterables.reverse(iterable),
        predicate=predicate,
    )

# ....................{ GETTERS ~ str                     }....................
#FIXME: Exercise both the "item_key" and "is_defaultable" parameters with
#exhaustive unit tests. This function has become astonishingly complex.
@type_check
def get_item_str_uniquified(
    # Mandatory parameters.
    iterable: IterableTypes,
    item_str_format: str,

    # Mutually exclusive parameters.
    item_attr_name: StrOrNoneTypes = None,
    item_key: StrOrNoneTypes = None,
    is_defaultable: bool = False,
) -> str:
    '''
    Create and return a new machine-readable string guaranteed to be unique
    across each attribute with the passed name of each item of the
    passed iterable if the ``item_attr_name`` parameter is non-``None`` *or*
    across each dictionary key with the passed name of each item of the passed
    iterable if the ``item_key`` parameter is non-``None``.

    Design
    ----------
    This function enforces uniqueness of either an object attribute *or*
    dictionary key assumed to be used by the caller as an SQL-like primary key
    uniquely identifying each item of this iterable. Specifically, this
    function:

    * If the ``item_attr_name`` parameter is non-``None``, requires each item
      of this iterable to declare an attribute with the passed name whose value
      is an arbitrary string.
    * If the ``item_key`` parameter is non-``None``, requires each item of this
      iterable to be a dictionary defining a key with the passed name whose
      value is an arbitrary string.

    Exactly one of the mutually exclusive ``item_attr_name`` and ``item_key``
    parameters *must* be passed. If neither *or* both of these parameters are
    passed, an exception is raised.

    This function then synthesizes a string suitable for use by callers as the
    value of that variable or key for a currently non-existing item of this
    iterable assumed to be created by the caller after calling this function.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable to be inspected.
    item_str_format : str
        Format specifier containing a ``{}`` substring (e.g., ``Item ({}).``),
        iteratively interpolated by this function with an arbitrary integer to
        produce the returned string.
    item_attr_name : StrOrNoneTypes
        Name of the attribute declared by *all* items of this list whose string
        values are to be uniquified. Defaults to ``None``, in which case the
        optional ``item_key`` parameter *must* be non-``None``.
    item_key : StrOrNoneTypes
        Key declared by *all* dictionary items of this list whose string values
        are to be uniquified. Defaults to ``None``, in which case the optional
        ``item_attr_name`` parameter *must* be non-``None``.
    is_defaultable : bool
        ``True`` only if this function implicitly accepts **non-compliant
        items** (i.e., items *not* declaring either an attribute with the name
        ``item_attr_name`` if that parameter is non-``None`` *or* the key
        ``item_key`` if that parameter is non-``None``). Specifically, if this
        boolean is:

        * ``True``, this function silently synthesizes a default string value
          for each non-compliant item of this iterable (e.g.,
          ``item_str_format.format(item_index)``, where ``item_index`` is the
          0-based index of that item in this iterable).
        * ``False``, this function raises an exception on visiting the first
          non-compliant item of this iterable.

        Defaults to ``False``.

    Returns
    ----------
    str
        New machine-readable string guaranteed to both match this format *and*
        be unique across all attributes with this name of all items of
        this iterable.

    Raises
    ----------
    BetseException
        If either:

        * The ``item_attr_name`` parameter is non-``None`` *and*, for one or
          more items of this iterable, either:

          * The ``is_defaultable`` parameter is ``False`` *and* this
            item contains no attribute with this name.
          * The value of this attribute in this item is *not* a string.

        * The ``item_key`` parameter is non-``None`` *and*, for one or more
          items of this iterable, either:

          * This item is *not* a dictionary.
          * The ``is_defaultable`` parameter is ``False`` *and* this
            dictionary does *not* contain this key.
          * The value of this key in this dictionary is *not* a string.
    BetseParamException
        If either:

        * Both of the ``item_attr_name`` *and* ``item_key`` parameters are
          passed.
        * Neither the ``item_attr_name`` *nor* ``item_key`` parameters are
          passed.
    BetseStrException
        If the passed format specifier contains no ``{}`` substring.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable.mapping import mappings
    from betse.util.type.obj import objects
    from betse.util.type.text.string import strs

    # Log this formatting.
    logs.log_debug(
        'Uniquifying iterable item variable "%s" with template "%s"...',
        item_attr_name, item_str_format)

    #FIXME: Ideally, we would also raise exceptions if this format
    #specifier contains two or more ``{}`` substrings. Sadly, there appears
    #no trivial and/or efficient means of doing so. While we suppose we
    #*COULD* define a new
    #betse.util.type.text.string.strs.get_substrs_count() function doing
    #so, the lack of any benefit to doing so hardly seems worth it. Note
    #also that genuinely testing this is effectively infeasible due to
    #false positives (e.g., the string "{}{}" should raise exceptions but
    #the string "{}{{}}" should *NOT*).

    # If this specifier contains no "{}" substring, raise an exception.
    #
    # Note that this simplistic logic fails to account for "{{" and "}}"
    # escaping and hence *COULD* fail to raise exceptions when passed
    # worst-case format specifiers, but that we mostly do not care.
    strs.die_unless_substr(text=item_str_format, substr='{}')

    # Callable accepting an arbitrary item and 0-based index of this item in
    # this iterable and returning the string value of the desired attribute or
    # dictionary key from this item. The signature of this callable resembles:
    #     def item_attr_getter(item: object, item_index: int) -> str
    item_str_getter = None

    # If uniquifying an attribute of these items...
    if item_attr_name is not None:
        # If uniquifying a dictionary key of these items, raise an exception.
        # There can only be one true item query paradigm.
        if item_key is not None:
            raise BetseParamException(
                '"item_attr_name" and "item_key" parameters both passed.')

        # If this function implicitly accepts non-compliant items, define a
        # lambda retrieving this string attribute from the passed item in a
        # manner defaulting to a string formatted with this item's index.
        if is_defaultable:
            item_str_getter = lambda item, item_index: (
                objects.get_attr_or_default(
                    obj=item,
                    attr_name=item_attr_name,
                    attr_default=item_str_format.format(item_index),
                    attr_type=str,
                ))
        # Else, this function explicitly rejects non-compliant items. In this
        # case, define a lambda retrieving this string attribute from the
        # passed item *WITHOUT* defaulting if this attribute is undefined.
        else:
            item_str_getter = lambda item, item_index: objects.get_attr(
                obj=item,
                attr_name=item_attr_name,
                attr_type=str,
            )
    # Else if uniquifying a dictionary key of these items...
    elif item_key is not None:
        # If this function implicitly accepts non-compliant items, define a
        # lambda retrieving this string value from the passed dictionary in a
        # manner defaulting to a string formatted with this dictionary's index.
        if is_defaultable:
            item_str_getter = lambda item, item_index: (
                mappings.get_key_value_or_default(
                    mapping=item,
                    key=item_key,
                    value_default=item_str_format.format(item_index),
                    value_type=str,
                ))
        # Else, this function explicitly rejects non-compliant items. In this
        # case, define a lambda retrieving this string value from the
        # passed dictionary *WITHOUT* defaulting if this key is undefined.
        else:
            item_str_getter = lambda item, item_index: mappings.get_key_value(
                mapping=item,
                key=item_key,
                value_type=str,
            )
    # Else, neither an attribute nor dictionary key of these items is
    # being uniquified. In this case, raise an exception.
    else:
        raise BetseParamException(
            '"item_attr_name" and "item_key" parameters not passed.')

    # Arbitrary integer to be formatted into the string to be returned. To
    # reduce the likelihood of collisions with existing items of this iterable,
    # this integer defaults to the 0-based index of the next item to add to
    # this iterable.
    #
    # Iteration below increments this integer until the string created by
    # formatting this integer into the passed format specifier produces a
    # string guaranteed to be unique across all items of this iterable.
    item_id = len(iterable) + 1

    # Uniquified string to be created and returned by this function.
    item_str = None

    # "True" only if the "item_str" string still collides with at least one
    # string value of an attribute with this name of an item of this
    # iterable and hence has yet to be uniquified.
    is_item_str_collides = True

    # Unordered set of all string values of each desired attribute or
    # dictionary key  of all items of this iterable if each item declares such
    # a variable or key *OR* raise an exception otherwise.
    item_strs = set(
        item_str_getter(item, item_index)
        for item_index, item in enumerate(iterable))

    # While the string value to be returned still collides with at least one
    # string value of an attribute with this name of an item of this
    # iterable and hence has yet to be uniquified...
    while is_item_str_collides:
        # Uniquified string to be created and returned by this function.
        item_str = item_str_format.format(item_id)

        # If this value collides with the uniquified string to be created
        # and returned by this function, this string is non-unique. In this
        # case, increment this identifier, format a new attribute with this
        # identifier, and continue searching.
        is_item_str_collides = item_str in item_strs

        # Unconditionally increment the integer to be formatted into the
        # "item_str" string by the next iteration of this loop.
        item_id += 1

    # Return this uniquified string.
    return item_str



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Class hierarchy designing **simulation phase requirements** (i.e., objects
encapsulating the current state of a single simulation feature for a given
simulation phase).
'''

# ....................{ IMPORTS                            }....................
from betse.science.phase.phasecls import SimPhase
from betse.science.phase.require.abc.phasereqabc import SimPhaseRequirementABC
# from betse.util.io.log import logs
from betse.util.type.decorator.decmemo import property_cached
from betse.util.type.iterable import itertest
from betse.util.type.iterable.set.setcls import FrozenSetSubclassable
from betse.util.type.text.string import strjoin
from betse.util.type.types import type_check, IterableOrNoneTypes, NoneType

# ....................{ SUBCLASSES ~ requirements          }....................
class SimPhaseRequirements(SimPhaseRequirementABC, FrozenSetSubclassable):
    '''
    Immutable set of simulation phase requirements, requiring zero or more
    arbitrary requirements to be satisfied.

    Operators
    ---------
    This set strictly conforms to the :class:`frozenset` API and hence supports
    the following set operators for combining multiple instances of this class:

    * ``|``, performing set union.
    * ``&``, performing set intersection.
    * ``-``, performing asymmetric set difference.
    * ``^``, performing symmetric set difference.

    Likewise, this set supports the following set operators for comparing any
    two instances of this class in terms of subset and superset relations:

    * ``a < b``, ``True`` only if set ``a`` is a proper subset of set ``b``.
    * ``a <= b``, ``True`` only if set ``a`` is a subset of set ``b``.
    * ``a > b``, ``True`` only if set ``a`` is a proper superset of set ``b``.
    * ``a >= b``, ``True`` only if set ``a`` is a superset of set ``b``.

    Design
    ----------
    This set satisfies the :class:`SimPhaseRequirementABC` API and is thus
    itself a valid requirement. Indeed, this set permits instances of that
    lower-level API to be efficiently composed together via an immutable
    set-like API and is thus recommended over that API for external usage.

    This set inherits the abstract :class:`collections.abc.Set` mixin rather
    than the concrete :class:`frozenset` type, as the latter is *not* a mixin.
    In all respects, however, this set is usable as a :class:`frozenset`.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __new__(
        cls, iterable: IterableOrNoneTypes = None) -> 'SimPhaseRequirements':
        '''
        Create, initialize, and return this requirement.

        Design
        ------
        To satisfy the :class:`frozenset` API, this method *must* accept only
        a single iterable. To quote `official documentation`_:

            Since some set operations create new sets, the default mixin
            methods need a way to create new instances from an iterable. The
            class constructor is assumed to have a signature in the form
            ``ClassName(iterable)``. That assumption is factored-out to an
            internal classmethod called :meth:`_from_iterable` which calls
            ``cls(iterable)`` to produce a new set. If the :class:`Set` mixin
            is being used in a class with a different constructor signature,
            you will need to override :meth:`_from_iterable` with a classmethod
            that can construct new instances from an iterable argument.

        .. _official documentation:
            https://docs.python.org/3/library/collections.abc.html#collections.abc.AsyncGenerator

        Parameters
        ----------
        iterable : IterableOrNoneTypes
            Iterable of zero or more arbitrary requirements, collectively
            defining this requirement. Since this requirement is an immutable
            set, this iterable is the *only* means of defining this
            requirement. Defaults to ``None``, in which case this set is
            permanently empty.
        '''

        # If no iterable was passed, default this iterable to the empty tuple.
        if iterable is None:
            iterable = ()

        # If any item of the passed iterable is *NOT* a requirement, raise an
        # exception.
        itertest.die_unless_items_instance_of(
            iterable=iterable, cls=SimPhaseRequirementABC)

        # Create, initialize, and return a new instance of this class
        # containing all unique items of this iterable.
        return super().__new__(cls, iterable)

        # self = super().__new__(cls, iterable)
        # logs.log_info('requirements for "%s": %r', self.name, self)
        # return self

    # ..................{ SUPERCLASS ~ requirement           }..................
    # Abstract properties required to be implemented by the
    # "SimPhaseRequirementABC" superclass.

    @property_cached
    def name(self) -> str:
        '''
        Human-readable name of this requirement, defined as the conjunction of
        the names of all requirements comprising this set (e.g., "full solver,
        fluid flow, and calcium ions (Ca2+)").

        For usability, the first character of this name should typically be
        lower- rather than uppercase.
        '''

        # If this set is empty, return a sane empty name.
        if not self:
            return 'no simulation features'
        # Else, this set contains at least one requirement.

        # Generator yielding the name of each requirement in this set.
        requirement_names = (requirement.name for requirement in self)

        # Create, return, and cache this name to be a human-readable
        # conjunction of these names.
        return strjoin.join_as_conjunction(*requirement_names)


    @type_check
    def is_satisfied(self, phase: SimPhase) -> bool:
        '''
        ``True`` only if all child requirements composed by this parent
        requirement are satisfied by the passed simulation phase.
        '''

        # This requirement is satisfied if and only if...
        return all(
            # This child requirement is satisfied by this phase.
            requirement.is_satisfied(phase)
            # For each child requirement of this parent requirement...
            for requirement in self)


    @type_check
    def set_satisfied(self, phase: SimPhase) -> None:
        '''
        Modify the passed simulation phase as needed to ensure that all child
        requirements composed by this parent requirement are satisfied by this
        phase.
        '''

        # For each child requirement of this parent requirement...
        for requirement in self:
            # Ensure this child requirement is satisfied by this phase.
            requirement.set_satisfied(phase)

    # ..................{ TESTERS                            }..................
    # For orthogonality with the existing frozenset.isdisjoint() method, this
    # method is intentionally *NOT* named is_intersection().
    def isintersection(self, other: object) -> bool:
        '''
        ``True`` only if the passed object is also an immutable set of
        simulation phase requirements *and* these two sets are **intersecting**
        (i.e., these sets have a non-null intersection), in which case these
        sets share one or more requirements in common.
        '''

        return not self.isdisjoint(other)

# ....................{ TYPES                              }....................
SimPhaseRequirementsOrNoneTypes = (SimPhaseRequirements, NoneType)
'''
Tuple of both the immutable requirements set tupe *and* that of the singleton
``None`` object.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **command** (i.e., external executable file) facilities.
'''

# ....................{ IMPORTS                           }....................
import sys
from betse.exceptions import BetseCommandException
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.types import type_check, StrOrNoneTypes

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_unless_command(filename: str, reason: StrOrNoneTypes = None) -> None:
    '''
    Raise an exception unless a command with the passed filename exists.

    Parameters
    ----------
    filename : str
        Either the basename *or* the absolute or relative filename of the
        executable file to be validated.
    reason : optional[str]
        Human-readable sentence fragment to be embedded in this exception's
        message (e.g., ``due to "pyside2-tools" not being installed``).
        Defaults to ``None``, in which case this message has no such reason.

    Raises
    ----------
    BetseCommandException
        If this command does *not* exist.

    See Also
    ----------
    :func:`is_command`
        Further details.
    '''

    # If this command does *NOT* exist...
    if not is_command(filename):
        # Exception message to be raised.
        message = 'Command "{}" not found'.format(filename)

        # If an exception reason was passed, embed this reason in this message.
        if reason is not None:
            message += ' ({})'.format(reason)

        # Finalize this message.
        message += '.'

        # Raise this exception.
        raise BetseCommandException(message)

# ....................{ TESTERS                           }....................
@type_check
def is_command(filename: str) -> bool:
    '''
    ``True`` only if a command with the passed filename exists.

    This is the case if this path is either:

    * The basename of an executable file in the current ``${PATH}``.
    * The relative or absolute path of an executable file.

    Parameters
    ----------
    pathname : str
        Either the basename *or* the absolute or relative filename of the
        executable file to be tested.

    Returns
    ----------
    bool
        ``True`` only if this command exists.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import files, pathnames
    from betse.util.os.command import cmdpath

    # This path is that of an existing command if and only if either...
    return (
        # This path is that of an executable file *OR*
        files.is_executable(filename) or (
        # This path is that of a basename in the current ${PATH}.
        pathnames.is_basename(filename) and cmdpath.is_pathable(filename))
    )

# ....................{ GETTERS                           }....................
@func_cached
def get_current_basename() -> str:
    '''
    Basename of the command originating the active Python interpreter.

    If this interpreter is interpreting a block of arbitrary runtime code
    passed to this interpreter on the command line via Python's ``-c`` option
    (e.g., due to being called by a distributed ``pytest-xdist`` test), this
    function unconditionally returns the basename of the current application
    (e.g., ``betse``) rather than ``-c``. Why? We can all agree that ``-c`` is
    an unexpected and rather non-human-readable basename, which is bad.
    '''

    # Avoid circular import dependencies.
    from betse.util.app.meta import appmetaone
    from betse.util.path import pathnames

    # Raw absolute or relative path of the current command.
    current_basename = sys.argv[0]

    # If this is the non-human-readable "-c" Python interpreter option,
    # substitute this with the human-readable basename of this application.
    if current_basename == '-c':
        current_basename = appmetaone.get_app_meta().package_name
    # Else, reduce this absolute or relative path to a basename.
    else:
        current_basename = pathnames.get_basename(current_basename)

    # Return this basename.
    return current_basename



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Classes to easily run BETSE simulations from an external script.
'''

# ....................{ IMPORTS                            }....................
import os
import numpy as np
from beartype import beartype
from beartype.typing import Optional
from betse.science import filehandling as fh
from betse.science.enum.enumphase import SimPhaseKind
from betse.science.parameters import Parameters as p
from betse.science.phase.phasecls import SimPhase
from betse.science.simrunner import SimRunner
from betse.lib import libs
from betse.lib.numpy.npcsv import write_csv
from betse.lib.pil import pilnumpy
from betse.lib.pil.pilnumpy import ImageModeType
from betse.science.chemistry.netplot import plot_master_network
from betse.science.math import finitediff as fd
from betse.util.io.log import logs
from betse.util.io.log.conf import logconf
from betse.util.io.log.logenum import LogLevel
from betse.util.math.mathinterp import interp2d_linear
from betse.util.path import files

# ....................{ CLASSES                            }....................
class BetseWrapper(object):
    '''
    Object allowing for simple creation of BETSE cell cluster and simulation
    object that can be easily worked with in an external script using BETSE as a
    dependency.

    This class creates (or optionally loads) a BETSE cells object, runs (or
    optionally loads) an init phase, and optionally runs a sim phase.
    '''

    # ..................{ INITIALIZERS                       }..................
    @beartype
    def __init__(
        self,

        # Mandatory parameters.
        config_filename: str,

        # Optional parameters.
        log_filename: Optional[str] = None,

        #FIXME: Validate with something resembling:
        #    log_level: Optional[Union[Literal['ALL'], ..., Literal['None']]] = None,
        log_level: Optional[str] = None,
    ) -> None:
        '''
        Initialization routines for the BETSE Wrapper.

        Parameters
        ----------
        config_filename : str
            Path to the config filename to run the betse simulation.
        log_filename : Optional[str]
            Absolute or relative filename to log simulation messages into.
            Defaults to ``None``, in which case the default filename is
            used.
        log_level : Optional[str]
            Valid options are:

            * 'ALL'.
            * 'DEBUG'.
            * 'INFO'.
            * 'WARNING'.
            * 'ERROR'.
            * 'CRITICAL'.
            * 'NONE'.

            String applying custom settings to the logging level of the
            BetseWrapper. If the log_level is specified by one of the above
            strings, then it overrides the verbose key in BetseWrapper methods
            with the specified logging function.
        '''

        # Classify all passed parameters.
        self._config_filename = config_filename
        self._log_filename = log_filename

        if log_level is not None:
            self._log_level = getattr(logs.LogLevel, log_level)
        else:
            self._log_level = None

    # ..................{ RUNNERS                            }..................
    @beartype
    def run_pipeline(
        self,
        new_mesh: bool = True,
        verbose: bool = True,
        run_init: bool = True,
        run_sim: bool = False,
    ) -> None:
        '''
        Runs an entire BETSE modelling pipeline -- including creating or loading
        a cell cluster, running an init phase simulation, and running a sim
        phase simulation.

        Parameters:
        --------------
        config_filename : str
            Full path to the configuration file for the simulation.
        new_mesh : bool
            Whether to generate a whole new mesh (True) or try to load a saved
            one (False).
        verbose: bool
            Spit out comments (True) or stay silent (False).
        run_init : bool
            Whether to run through the BETSE initialization (True) or not
            (False).
        run_sim : bool
            whether to run a BETSE simulation phase (True) or not (False).
        '''

        # Make an instance of the BETSE 'parameters' object based on settings in
        # the configuration file supplied.
        self.p = p.make(self._config_filename)

        self._set_logging(verbose=verbose)

        # Make or load a BETSE cell cluster.
        self._make_mesh(new_mesh=new_mesh)

        if run_init:
            self._init_runner(runsim=run_sim)

        if self.verbose is True:
            logs.log_info("Successfully run betse pipeline!")


    @beartype
    def run_seed(self, verbose: bool = False):
        '''
        Initializes the BETSE modelling object, which includes creating
        a cell cluster.

        Parameters:
        --------------
        '''

        self.p = p.make(self._config_filename)

        self.verbose = verbose  # save verbosity setting

        self._set_logging(verbose=verbose)

        self._make_mesh(new_mesh=True)  # make a BETSE cell cluster

        if self.verbose is True:
            logs.log_info(
                "Successfully created a new BETSE cell cluster object!")


    @beartype
    def run_init(self, new_mesh: bool = True, verbose: bool = False):
        '''
        Initializes the BETSE modelling object, which includes creating or loading
        a cell cluster, and running an init phase simulation.

        Parameters:
        --------------
        new_mesh : bool
            Whether to generate a whole new mesh (True) or try to load a saved one (False).
        verbose: bool
            Spit out comments (True) or stay silent (False).
        '''

        # Make an instance of the BETSE 'parameters' object based on
        # settings in the configuration file supplied:

        self.p = p.make(self._config_filename)

        self.verbose = verbose  # save verbosity setting

        self._set_logging(verbose=verbose)

        self._make_mesh(new_mesh=new_mesh)  # make or load a BETSE cell cluster

        self._init_runner(runsim=False)

        if self.verbose is True:
            logs.log_info("Successfully run initialization on BETSE model!")


    @beartype
    def run_sim(self, verbose: bool = False):
        '''
        Loads a previously-made BETSE cell cluster and init phase simulation to run
        the sim phase simulation.

        Parameters:
        --------------
        verbose: bool
            Spit out comments (True) or stay silent (False).
        '''

        # Make an instance of the BETSE 'parameters' object based on
        # settings in the configuration file supplied:

        self.p = p.make(self._config_filename)

        self.verbose = verbose  # save verbosity setting

        self._set_logging(verbose=verbose)

        self._make_mesh(new_mesh=False)  # load a BETSE cell cluster

        self._sim_runner() # Run the BETSE simulation

        if self.verbose is True:
            logs.log_info("Successfully run simulation on BETSE model!")

    # ..................{ LOADERS                            }..................
    #FIXME: Docstring us up, please. Flying churros at midnight!
    @beartype
    def load_seed(self, verbose: bool = False):
        '''
        '''

        self.p = p.make(self._config_filename)

        self.verbose = verbose  # save verbosity setting

        self._set_logging(verbose=verbose)

        if not files.is_file(self.p.seed_pickle_filename):  # If file doesn't exist...
            if self.verbose is True:
                logs.log_warning("File not found. Run seed to create a cell cluster.")

        else:  # Otherwise, load the saved cell cluster:
            if self.verbose is True:
                logs.log_info("Loading 2D Grid from file.")

            # Load from previous creation:
            cells, _ = fh.loadWorld(self.p.seed_pickle_filename)

            self.simrun = SimRunner(self.p)

            # Simulation phase, created after unpickling these objects above
            self.phase = SimPhase(
                kind=SimPhaseKind.SEED,
                p=self.p,
                cells=cells,
            )


    @beartype
    def load_init(self, verbose: bool = False):
        '''
        '''

        self.p = p.make(self._config_filename)

        self.verbose = verbose  # save verbosity setting

        self._set_logging(verbose=verbose)

        if not files.is_file(self.p.init_pickle_filename):  # If file doesn't exist...
            if self.verbose is True:
                logs.log_warning("File not found. Run init to create an initialization.")

        else:  # Otherwise, load the saved cell cluster:
            if self.verbose is True:
                logs.log_info("Loading BETSE init from file.")

            sim, cells, _ = fh.loadSim(self.p.init_pickle_filename)

            self.simrun = SimRunner(self.p)

            # Simulation phase, created after unpickling these objects above
            self.phase = SimPhase(
                kind=SimPhaseKind.INIT,
                p=self.p,
                cells=cells,
                sim=sim,
            )

            if self.verbose is True:
                logs.log_info("Successfully loaded init of BETSE model!")


    @beartype
    def load_sim(self, verbose: bool = False):
        '''
        '''

        self.p = p.make(self._config_filename)

        self.verbose = verbose  # save verbosity setting

        self._set_logging(verbose=verbose)

        if not files.is_file(self.p.sim_pickle_filename):  # If file doesn't exist...
            if self.verbose is True:
                logs.log_warning("File not found. Run sim to create a simulation.")

        else:  # Otherwise, load the saved cell cluster:
            if self.verbose is True:
                logs.log_info("Loading BETSE sim from file.")

            sim, cells, _ = fh.loadSim(self.p.sim_pickle_filename)

            self.simrun = SimRunner(self.p)

            # Simulation phase, created after unpickling these objects above
            self.phase = SimPhase(
                kind=SimPhaseKind.SIM,
                p=self.p,
                cells=cells,
                sim=sim,
            )

            if self.verbose is True:
                logs.log_info("Successfully loaded sim of BETSE model!")


    @beartype
    def load_simgrn(self, verbose: bool = False):
        '''
        '''

        self.p = p.make(self._config_filename)

        self.verbose = verbose  # save verbosity setting

        self._set_logging(verbose=verbose)

        if not files.is_file(self.p.sim_pickle_filename):  # If file doesn't exist...
            if self.verbose is True:
                logs.log_warning("File not found. Run sim_grn to create a new grn simulation.")

        else:  # Otherwise, load the saved cell cluster:
            if self.verbose is True:
                logs.log_info("Loading simulated grn from file.")

            init, _, _ = fh.loadSim(self.p.init_pickle_filename)
            grn, cells, _ = fh.loadSim(self.p.grn_pickle_filename)

            # Simulation phase.
            self.phase = SimPhase(
                kind=SimPhaseKind.INIT,
                cells=cells,
                sim=init,
                p=self.p)

            self.phase.sim.grn.core = grn

            if self.verbose is True:
                logs.log_info(
                    "Successfully loaded simulated grn of BETSE model!")


    @beartype
    def run_sim_grn(self, new_mesh: bool = True, verbose: bool = False):
        '''
        Run only the BETSE GRN of the model (no bioelectricity).
        '''

        self.p = p.make(self._config_filename)

        self.verbose = verbose  # save verbosity setting

        self._set_logging(verbose=verbose)

        self._make_mesh(new_mesh=new_mesh)  # load a BETSE cell cluster

        self.phase = self.simrun.sim_grn()  # Run the BETSE simulation as GRN-only

        if self.verbose is True:
            logs.log_info(
                "Successfully run GRN-only simulation on BETSE model!")

    # ..................{ GETTERS                            }..................
    @beartype
    def get_network(self, verbose: bool = True, as_networkx: bool = False):
        '''
        Returns BETSE's GRN (if simulated) as a pydot (as_networkx=False) or
        networkx (as_networkx = True) Digraph object.
        '''

        if self.phase.sim.grn is not None:
            # Working with BETSE's networks:
            # Access the gene regulatory network core:
            grn = self.phase.sim.grn.core

            graph_net = plot_master_network(grn, self.p)

            if as_networkx is True:
                networkx = libs.import_runtime_optional('networkx')
                # Convert the pydot graph to a networkx file:
                graph_net = networkx.nx_pydot.from_pydot(graph_net)

        else:
            if verbose is True:
                logs.log_info("No GRN present in this BETSE model.")
            graph_net = None

        return graph_net


    @beartype
    def get_betse_grn(self, verbose: bool = True):
        '''
        Returns BETSE's GRN modelling object (if a GRN is simulated in the BETSE
        model).
        '''

        if self.phase.sim.grn is not None:
            # Working with BETSE's networks:
            # Access the gene regulatory network core:
            grn = self.phase.sim.grn.core

        else:
            if verbose is True:
                logs.log_info("No GRN present in this BETSE model.")
            grn = None

        return grn


    @beartype
    def get_connected_grn_elements(self, verbose: bool = False):
        '''
        Returns a list of sets of connected nodes of BETSE's GRN (if simulated).
        '''

        graph_network = self.get_network(verbose=verbose, as_networkx=True)

        if graph_network is not None:
            graph_network_o = graph_network.to_undirected()
            networkx = libs.import_runtime_optional('networkx')
            connected_elements = sorted(networkx.connected_components(graph_network_o))

            if verbose is True:
                print(connected_elements)

        else:
            if verbose is True:
                logs.log_info("No GRN present in this BETSE model.")
            connected_elements = None

        return connected_elements

    # ..................{ PLOTTERS                           }..................
    @beartype
    def plot_network(self, verbose: bool = True):
        '''
        Exports an svg of BETSE's GRN (if simulated).
        '''

        if self.phase.sim.grn is not None:
            # Working with BETSE's networks:
            # Access the gene regulatory network core:
            grn = self.phase.sim.grn.core

            graph_pydot = plot_master_network(grn, self.p)

            # Save the pydot graph to an svg file:
            # Initialize saving:
            grn.init_saving(self.phase.cells, self.p, plot_type='init', nested_folder_name='GRN')

            # Optionally print the location of the image path using: print(grn.imagePath)
            savename = os.path.join(grn.imagePath[0:-4], 'OptimizedNetworkGraph.svg')
            graph_pydot.write_svg(savename, prog='dot')

            if verbose is True:
                logs.log_info(f"Model GRN network image saved to {savename}.")

        else:
            if verbose is True:
                logs.log_info("No GRN present in this BETSE model.")

    # ..................{ OTHERS                             }..................
    @beartype
    def analyze_network(
        self,
        verbose: bool = True,
        plot_network: bool = False,
        save_csv: bool = True,
    ):
        '''
        Analyzes the node connectivity (node degree) of a BETSE GRN (if
        simulated) and optionally exports results to csv (save_csv = True). The
        graph of the GRN can also be optionally exported (plot_network=True).
        '''

        if self.phase.sim.grn is not None:
            # Import optional dependencies for working with networks:
            # Defer importation of optional runtime dependencies until necessary.
            pydot, networkx = libs.import_runtime_optional('pydot', 'networkx')

            # Working with BETSE's networks:
            # Access the gene regulatory network core:
            grn = self.phase.sim.grn.core

            graph_pydot = plot_master_network(grn, self.p)

            if plot_network is True:
                # Save the pydot graph to an svg file:
                # Initialize saving:
                grn.init_saving(self.phase.cells, self.p, plot_type='init', nested_folder_name='GRN')

                # Optionally print the location of the image path using: print(grn.imagePath)
                savename = os.path.join(grn.imagePath[0:-4], 'OptimizedNetworkGraph.svg')
                graph_pydot.write_svg(savename, prog='dot')

                if verbose is True:
                    logs.log_info(f"Model GRN network image saved to {savename}.")

            # Convert the pydot graph to a networkx file:
            graph_network = networkx.nx_pydot.from_pydot(graph_pydot)

            # Compute the connectivity degree of graph nodes:
            degree_list = np.asarray(sorted(networkx.degree(graph_network)))

            # determine the connectivity degree values as an array of floats:
            degree_vals = np.asarray(degree_list[:, 1], dtype=float)

            # get the indices into the degree list that will sort the connectivity from high to low:
            inds_sort = np.flip(np.argsort(degree_vals))

            # get the node names and degree values sorted from highest to lowest connectivity degree:
            sorted_nodes = degree_list[inds_sort, 0]
            sorted_vals = degree_list[inds_sort, 1]

            if verbose is True:
                for node, val in zip(sorted_nodes, sorted_vals):
                    logs.log_info(f'{repr(node)}, {repr(val)}')


            if save_csv is True:
                savename = os.path.join(grn.imagePath[0:-4], 'NetworkConnectivity.csv')
                column_name_to_values = {
                    'Node name': sorted_nodes,
                    'Connectivity': sorted_vals,
                }
                column_name_to_format = {
                    'Node name': '%s',
                    'Connectivity': '%s',
                }

                write_csv(
                    filename=savename,
                    column_name_to_values=column_name_to_values,
                    column_name_to_format=column_name_to_format,
                )

                if verbose is True:
                    logs.log_info(f"Analyzed BETSE model GRN network and exported csv to {savename}.")

        else:
            if verbose is True:
                logs.log_info("No GRN present in this BETSE model.")

        return sorted_nodes, sorted_vals


    @beartype
    def interp_bitmap_to_cells(
        self,
        bitmap_filename: str,
        to_mems: bool = True,
        smooth: bool = False,
    ):
        '''
        Interpolate a greyscale bitmap (supplied in the bitmap_filename string)
        to the cell cluster. The resulting interpolation is an array of floats,
        where 0.0 is black and 255.0 is white.

        Parameters
        -------------
        bitmap_filename : str
            Path and filename to the bitmap to read and interpolate as a
            greyscale image.
        to_mems : bool
            If True, interpolates the bitmap to membrane domains of the cell
            cluster. If False, interpolates the bitmap to cell centres of the
            cell cluster.
        smooth : bool
            Apply smoothing to the interpolated bitmap (True) or keep raw
            interpolation (False).

        Returns
        -------
        ndarray
            An array (shaped to membrane domains or for to_mems == True and cell
            centres for to_mems=False) of bitmap greyscale values interpolated
            over the cell cluster. Values of
            0.0 are pure black while 255.0 are pure white.
        '''

        cells = self.phase.cells

        xmi = cells.xmin
        ymi = cells.ymin
        xma = cells.xmax
        yma = cells.ymax

        if to_mems:
            xmap = cells.map_mem2ecm

        else:
            xmap = cells.map_cell2ecm

        xx = np.linspace(cells.xmin, cells.xmax, cells.X.shape[1])
        yy = np.linspace(cells.ymin, cells.ymax, cells.X.shape[0])

        # Three-dimensional Numpy array of the RGB-ordered integer components of all
        # pixels loaded from this image.
        a1o = pilnumpy.load_image(filename=bitmap_filename, mode=ImageModeType.GRAYSCALE_FLOAT)

        a1 = np.asarray(a1o, dtype=np.float64)

        a1 = np.flipud(a1)

        xa = np.linspace(xmi, xma, a1.shape[1])
        ya = np.linspace(ymi, yma, a1.shape[0])

        spline_F = interp2d_linear(xa, ya, a1)
        # spline_F = interpolate.interp2d(xa, ya, a1, kind='linear', fill_value=0.0)
        fe = spline_F(xx, yy)

        if smooth:
            fe = fd.integrator(fe, sharp=0.5)  # smooth a little to avoid bizarre visual effects

        f = fe.ravel()[xmap]

        return f

    # ..................{ PRIVATE                            }..................
    @beartype
    def _set_logging(self, verbose: bool = False) -> None:
        '''
        Set the logging properties of the BetseWrapper.
        '''

        self.verbose = verbose  # save verbosity setting

        # Logging configuration singleton.
        log_config = logconf.get_log_conf()

        # If the user passed a log filename, reconfigure our logging
        # configuration to log to this file.
        if self._log_filename is not None:
            log_config.filename = self._log_filename
        # Else, the user passed *NO* log filename. In this case, accept the
        # current default log filename.

        # If the user passed *NO* log level, default to a log level
        # corresponding to the passed verbosity.
        if self._log_level is None:
            if verbose:
                log_config.handler_stdout.setLevel(LogLevel.INFO)
            else:
                log_config.handler_stdout.setLevel(LogLevel.WARNING)
        # Else, the user has set a log level. In this case, apply it.
        else:
            log_config.handler_stdout.setLevel(self._log_level)


    @beartype
    def _make_mesh(self, new_mesh: bool = False):
        '''
        Generates or loads a saved 2D Voronoi mesh for the BETSE simulation,
        based on settings in the supplied config file.

        Parameters
        ----------
        new_mesh : bool
            Whether to generate a whole new mesh (True) or try to load a saved
            one (False).
        '''

        if new_mesh is True:  # If 'new mesh' is requested, make a whole new cell cluster

            if self.verbose is True:
                logs.log_info("Creating a new cell cluster.")

            # Create a new grid:

            self.simrun = SimRunner(self.p)  # Call an instance of BETSE's "SimRunner'
            phase = self.simrun.seed()  # Go through the making of a whole cell cluster
            self.phase = phase  # save the phase object (which contains the cell cluster
            # to the DemoSim object)

        else:  # otherwise, if a new mesh isn't needed, go ahead and load a saved cluster,
            # if it exists!

            if not files.is_file(self.p.seed_pickle_filename):  # If it doesn't exist...
                if self.verbose is True:
                    logs.log_warning("File not found; Creating a new cell cluster...")

                # Make a new mesh
                self.simrun = SimRunner(self.p)
                phase = self.simrun.seed()  # Go ahead and make a new cluster
                self.phase = phase

            else:  # Otherwise, load the saved cell cluster:
                if self.verbose is True:
                    logs.log_info("Loading a cell cluster from file.")

                # Load from previous creation:
                cells, _ = fh.loadWorld(self.p.seed_pickle_filename)

                self.simrun = SimRunner(self.p)

                # Simulation phase, created after unpickling these objects above

                self.phase = SimPhase(
                    kind=SimPhaseKind.SEED,
                    p=self.p,
                    cells=cells,
                )

                # Initialize core simulation data structures.

                self.phase.sim.init_core(self.phase)
                self.phase.dyna.init_profiles(self.phase)

        # assign commonly used dimensional parameters to the DemoSim object:
        self._assign_shorts(self.phase.cells)


    @beartype
    def _init_runner(self, runsim: bool = False):
        '''
        Run the BETSE cell cluster through the initialization specified in
        the loaded configuration file.

        Parameters
        ----------
        runsim : bool
        '''

        self.phase = self.simrun.init()

        if runsim:
            self.phase = self.simrun.sim()

        # short forms of commonly used data structures from BETSE sim object, based on
        # ion settings for 'basic' profile:

        # re-assign commonly used dimensional parameters:
        self._assign_shorts(self.phase.cells)


    def _sim_runner(self):
        '''
        Run the BETSE cell cluster through the initialization specified in
        the loaded configuration file.

        Parameters
        ----------
        runsim : bool
        '''

        self.phase = self.simrun.sim()

        # short forms of commonly used data structures from BETSE sim object, based on
        # ion settings for 'basic' profile:

        # re-assign commonly used dimensional parameters:
        self._assign_shorts(self.phase.cells)


    def _assign_shorts(self, cells):
        '''
        Assigns short forms to commonly used mesh components.

        Parameters
        -----------
        cells : ???
            BETSE cell cluster object.
        '''

        # Data defined on cell centrepoints (self.xc, self.yc):
        self.vm_ave = self.phase.sim.vm_ave  # Vmem state averaged over a whole cell [V]
        self.cc_cells = self.phase.sim.cc_cells  # Array of cytosolic ion conc arrays [mol/m^3]
        self.Ex = self.phase.sim.E_cell_x  # x-component of electric field [V/m]
        self.Ey = self.phase.sim.E_cell_y  # y-component of electric field [V/m]
        self.rho_cells = self.phase.sim.rho_cells  # charge density [C/m^3]

        # Data defined on membrane midpoints (self.xm, self.ym):

        self.vm = self.phase.sim.vm  # Vmem defined at cell edges (membranes) [V]
        self.Dm_cells = self.phase.sim.Dm_cells  # Array of mem diffusion constant arrays [m^2/s]

        # Data defined on environmental points (self.xenv, self.yenv)

        self.cc_env = self.phase.sim.cc_env  # Array of extracellular ion conc arrays [mol/m^3]

        # Single-value data (one value for each ion included in the simulation)

        self.D_free = self.phase.sim.D_free  # Array of media diffusion constant arrays [m^2/s]

        # Numerical indices to access specific ions in cc_cells, Dm_cells, cc_env, etc

        self.iNa = self.phase.sim.iNa  # Index of Na ion in array-of-arrays
        self.iK = self.phase.sim.iK  # Index of K ion in array-of-arrays
        self.iM = self.phase.sim.iM  # Index of M ion in array-of-arrays
        self.iP = self.phase.sim.iP  # Index of P ion in array-of-arrays


        # points of cell centres:

        self.xc = cells.cell_centres[:, 0]
        self.yc = cells.cell_centres[:, 1]

        # membrane midpoints for edges of each cell:

        self.xmem = cells.mem_mids_flat[:, 0]
        self.ymem = cells.mem_mids_flat[:, 1]

        # points of the square environmental grid:

        self.xenv = cells.xypts[:, 0]
        self.yenv = cells.xypts[:, 1]

        # global size limits of the whole environment and cell cluster:

        self.xyaxis = np.asarray([cells.xmin, cells.xmax, cells.ymin, cells.ymax])

        # vertices of each polygon representing each cell in the cluster:
        self.verts = cells.cell_verts

        # Normal vectors to membrane edges
        self.nx = cells.mem_vects_flat[:, 2]
        self.ny = cells.mem_vects_flat[:, 3]

        # true extracellular matrix points (point shared between two membrane mids):
        self.xec = cells.ecm_mids[:, 0]
        self.yec = cells.ecm_mids[:, 1]

        # assign length of cell (cdl), mems (mdl) and env (edl) of mesh:

        self.cdl = len(cells.cell_i)
        self.mdl = len(cells.mem_i)
        self.edl = len(cells.ecm_mids)
        self.envdl = len(cells.xypts)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Simulation export pipeline** (i.e., container of related, albeit isolated,
simulation export actions to be run iteratively) hierarchy.
'''

# ....................{ IMPORTS                           }....................
from betse.science.pipe.pipeabc import SimPipeABC
from betse.util.type.descriptor.descs import classproperty_readonly

# ....................{ SUPERCLASSES                      }....................
class SimPipeExportABC(SimPipeABC):
    '''
    Abstract base class of all **simulation export pipeline** (i.e., object
    iteratively exporting *all* variations on a single type of simulation
    export) subclasses.
    '''

    # ..................{ SUPERCLASS                        }..................
    @classproperty_readonly
    def _RUNNER_METHOD_NAME_PREFIX(self) -> str:
        return  'export_'


    @classproperty_readonly
    def _VERB_CONTINUOUS(cls) -> str:
        return 'Exporting'



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Voltage-gated funny current channel classes.
'''

# .................... IMPORTS                            ....................
from abc import ABCMeta, abstractmethod
import numpy as np
from betse.science import sim_toolbox as stb
from betse.science.channels.channelsabc import ChannelsABC
from betse.science.math import toolbox as tb
from betse.util.io.log import logs

# .................... BASE                               ....................
class VgFunABC(ChannelsABC, metaclass=ABCMeta):
    '''
    Abstract base class of all Voltage-gated funny current channel classes.

    Attributes
    ----------
    _mInf :
        Equillibrium value function for m-gates of channel
    _mTau :
        Time-constant function for m-gates of channel
    _hInf :
        Equillibrium value function for h-gates of channel
    _hTau :
        Time-constant function for h-gates of channel

    '''

    def init(self, vm, cells, p, targets = None):
        '''
        Initialize targeted voltage-gated funny current channels at the initial
        time step of the simulation based on the initial cell Vmems.

        Channel model uses Hodgkin-Huxley kinetic model style
        for voltage gated channels.
        '''

        if targets is None:

            self.targets = None

            if type(vm) == float or type(vm) == np.float64:
                self.data_length = 1

            else:
                self.data_length = len(vm)


        else:
            self.targets = targets
            self.data_length = len(self.targets)
            self.mdl = len(cells.mem_i)

        self.modulator = 1.0

        self.v_corr = 0.0  # in experiments, the measurement junction voltage is about 10 mV

        if self.targets is None:

            V = vm * 1000 + self.v_corr

        else:
            V = vm[self.targets] * 1000 + self.v_corr

        self._init_state(V)

        self.ions = ['Na', 'K', 'Ca']
        self.rel_perm = [self._PmNa, 1.0, self._PmCa]


    def run(self, vm, p):
        '''
        Handle all targeted voltage-gated funny current channels by working with the passed
        user-specified parameters on the tissue simulation and cellular
        world for a time step.

        Channel model uses Hodgkin-Huxley kinetic model style
        for voltage gated channels.

        '''

        if self.targets is None:

            V = vm*1000 + self.v_corr

        else:

            V = vm[self.targets] * 1000 + self.v_corr

        self._calculate_state(V)

        self._implement_state(V, p)

    def _implement_state(self, V, p):

        # Update the 'm' and 'h' gating functions:
        self.update_mh(p, time_unit = self.time_unit)

        # calculate the open-probability of the channel:
        P = (self.m ** self._mpower) * (self.h ** self._hpower)

        if self.targets is None:

            self.P = P

        else:
            self.P = np.zeros(self.mdl)
            self.P[self.targets] = P


    @abstractmethod
    def _init_state(self, V):
        '''
        Initializes values of the m and h gates of the channel.
        '''
        pass


    @abstractmethod
    def _calculate_state(self, V):
        '''
        Calculates time-dependent values of the m and h gates of the channel.
        '''
        pass

# ....................{ SUBCLASS                           }....................

class HCN2(VgFunABC):
    '''
    HCN2 model from 21 day old dorsal root ganglion of mouse. HCN channels are voltage-gated ionic channels,
    regulated by cyclic nucleotides, such as cyclic adenosine-mono-phosphate (cAMP) (not modelled here).
    In contrast to most Na+ and K+ ionic channels, which open when membrane potential is depolarized,
    they are opened when the membrane potential hyperpolarizes below -50 mV.

    This channel is expressed in brain and heart tissue, though the specific function is unknown.

    Reference:  Moosmang S. et al. Cellular expression and functional characterization of four
    hyperpolarization-activated pacemaker channels in cardiac and neuronal tissues.
    Eur. J. Biochem., 2001 Mar , 268 (1646-52).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the funny current channel: HCN2')

        self.time_unit = 1.0e3

        self.v_corr = 0

        V = V - 10

        # initialize values of the m and h gates of the HCN2 based on m_inf and h_inf:
        self.m = 1.0000 / (1 + np.exp((V - -99) / 6.2))
        self.h = 1

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 0

        self._PmCa = 0.05  # channel permeability to Ca2+
        self._PmNa = 0.2  # channel permeability ratio to Na+


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self.vrev = -45  # reversal voltage used in model [mV]

        V = V - 10

        self._mInf = 1.0000 / (1 + np.exp((V - -99) / 6.2))
        self._mTau = 184.0000
        self._hInf = 1
        self._hTau = 1

class HCN4(VgFunABC):
    '''
    HCN4 model from 21 day old dorsal root ganglion of mouse. HCN channels are voltage-gated ionic channels,
    regulated by cyclic nucleotides, such as cyclic adenosine-mono-phosphate (cAMP) (not modelled here).
    In contrast to most Na+ and K+ ionic channels, which open when membrane potential is depolarized,
    they are opened when the membrane potential hyperpolarizes below -50 mV.

    This channel is expressed in brain and heart tissue, and plays a crucial role in pacemaker activity of the
    heart, and possibly the repetitive firing of Purkunjie neurons.

    Reference:  Moosmang S. et al. Cellular expression and functional characterization of four
    hyperpolarization-activated pacemaker channels in cardiac and neuronal tissues.
    Eur. J. Biochem., 2001 Mar , 268 (1646-52).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the funny current channel: HCN4')

        self.time_unit = 1.0e3

        self.v_corr = 0

        V = V - 10

        # initialize values of the m and h gates of the HCN2 based on m_inf and h_inf:
        self.m = 1.0000 / (1 + np.exp((V - -100) / 9.6))
        self.h = 1

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 0

        self._PmCa = 0.05  # channel permeability to Ca2+
        self._PmNa = 0.2  # channel permeability ratio to Na+


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self.vrev = -45  # reversal voltage used in model [mV]

        V = V - 10

        self._mInf = 1.0000 / (1 + np.exp((V - -100) / 9.6))
        self._mTau = 461.0000
        self._hInf = 1
        self._hTau = 1

class HCN1(VgFunABC):

    """
    Fastest activation of the HCN channels, the HCN1 channel is also implicated in sinus rhythm control
    of the heart, brain activity, and in the retina.

    Reference: 3] 	Moosmang S. et al. Cellular expression and functional characterization of four
    hyperpolarization-activated pacemaker channels in cardiac and neuronal tissues.
    Eur. J. Biochem., 2001 Mar , 268 (1646-52).
    """

    def _init_state(self, V):

        logs.log_info('You are using the funny current channel: HCN1')

        self.time_unit = 1.0e3

        # initialize values of the m and h gates of the HCN2 based on m_inf and h_inf:
        self.m = 1.0000 / (1 + np.exp((V - -94) / 8.1))
        self.h = 1

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 0

        self._PmCa = 0.05  # channel permeability ratio to Ca2+
        self._PmNa = 0.2  # channel permeability ratio to Na+

    def _calculate_state(self, V):


        self.vrev = -45  # reversal voltage used in model [mV]

        self._mInf = 1.0000 / (1 + np.exp((V - -94) / 8.1))
        self._mTau = 30.0000
        self._hInf = 1
        self._hTau = 1

class HCNLeak(VgFunABC):
    '''
    HCN2 model from 21 day old dorsal root ganglion of mouse. HCN channels are voltage-gated ionic channels,
    regulated by cyclic nucleotides, such as cyclic adenosine-mono-phosphate (cAMP) (not modelled here).
    In contrast to most Na+ and K+ ionic channels, which open when membrane potential is depolarized,
    they are opened when the membrane potential hyperpolarizes below -50 mV.

    This channel is expressed in brain and heart tissue, though the specific function is unknown.

    Reference:  Moosmang S. et al. Cellular expression and functional characterization of four
    hyperpolarization-activated pacemaker channels in cardiac and neuronal tissues.
    Eur. J. Biochem., 2001 Mar , 268 (1646-52).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the funny current channel: HCN Leak')

        self.time_unit = 1.0

        self.v_corr = 0

        # initialize values of the m and h gates of the HCN2 based on m_inf and h_inf:
        self.m = np.ones(self.data_length)
        self.h = np.ones(self.data_length)

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 0
        self._hpower = 0

        self._PmCa = 0.05  # channel permeability to Ca2+
        self._PmNa = 0.33  # channel permeability ratio to Na+


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self.vrev = -45  # reversal voltage used in model [mV]

        self._mInf = 1.0
        self._mTau = 1.0
        self._hInf = 1.0
        self._hTau = 1.0

class HCN2_cAMP(VgFunABC):
    '''
    HCN2 model from 21 day old dorsal root ganglion of mouse. HCN channels are voltage-gated ionic channels,
    regulated by cyclic nucleotides, such as cyclic adenosine-mono-phosphate (cAMP), modelled here as a simple
    shift in the channel voltage gating threshold.
    In contrast to most Na+ and K+ ionic channels, which open when membrane potential is depolarized,
    they are opened when the membrane potential hyperpolarizes below -50 mV.

    This channel is expressed in brain and heart tissue, though the specific function is unknown.

    Reference:  Moosmang S. et al. Cellular expression and functional characterization of four
    hyperpolarization-activated pacemaker channels in cardiac and neuronal tissues.
    Eur. J. Biochem., 2001 Mar , 268 (1646-52).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the funny current channel: HCN2 with cAMP activation.')

        self.time_unit = 1.0e3

        self.v_corr = 0

        V = V - 20

        # initialize values of the m and h gates of the HCN2 based on m_inf and h_inf:
        self.m = 1.0000 / (1 + np.exp((V + 99) / 6.2))
        self.h = 1

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 0

        self._PmCa = 0.05  # channel permeability to Ca2+
        self._PmNa = 0.2  # channel permeability ratio to Na+


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self.vrev = -45  # reversal voltage used in model [mV]

        V = V - 20

        self._mInf = 1.0000 / (1 + np.exp((V + 99) / 6.2))
        self._mTau = 184.0000
        self._hInf = 1
        self._hTau = 1

class HCN4_cAMP(VgFunABC):
    '''
    HCN4 model from 21 day old dorsal root ganglion of mouse. HCN channels are voltage-gated ionic channels,
    regulated by cyclic nucleotides, such as cyclic adenosine-mono-phosphate (cAMP), modelled here as a simple
    shift in the channel voltage gating threshold.
    In contrast to most Na+ and K+ ionic channels, which open when membrane potential is depolarized,
    they are opened when the membrane potential hyperpolarizes below -50 mV.

    This channel is expressed in brain and heart tissue, and plays a crucial role in pacemaker activity of the
    heart, and possibly the repetitive firing of Purkunjie neurons.

    Reference:  Moosmang S. et al. Cellular expression and functional characterization of four
    hyperpolarization-activated pacemaker channels in cardiac and neuronal tissues.
    Eur. J. Biochem., 2001 Mar , 268 (1646-52).

    '''

    def _init_state(self, V):
        """

        Run initialization calculation for m and h gates of the channel at starting Vmem value.

        """

        logs.log_info('You are using the funny current channel: HCN4 with cAMP activation.')

        self.time_unit = 1.0e3

        self.v_corr = 0

        V = V - 20

        # initialize values of the m and h gates of the HCN2 based on m_inf and h_inf:
        self.m = 1.0000 / (1 + np.exp((V  + 100) / 9.6))
        self.h = 1

        # define the power of m and h gates used in the final channel state equation:
        self._mpower = 1
        self._hpower = 0

        self._PmCa = 0.05  # channel permeability to Ca2+
        self._PmNa = 0.20  # channel permeability ratio to Na+


    def _calculate_state(self, V):
        """

        Update the state of m and h gates of the channel given their present value and present
        simulation Vmem.

        """

        self.vrev = -45  # reversal voltage used in model [mV]

        V = V - 24

        self._mInf = 1.0000 / (1 + np.exp((V + 100) / 9.6))
        self._mTau = 461.0000
        self._hInf = 1
        self._hTau = 1




#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **bitwise** (i.e., callables operating on one or more individual bits
of integer-based bit fields) facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.util.type.types import type_check

# ....................{ TESTERS                           }....................
@type_check
def is_bit_on(bit_field: int, bit_mask: int) -> bool:
    '''
    ``True`` only if the bit in the passed bit field uniquely identified by the
    passed bit mask is **on** (i.e., one rather than zero).

    Parameters
    ----------
    bit_field : int
        **Bit field** (i.e., integer whose individual bits collectively
        comprise a rudimentary array of boolean flags) to be tested.
    bit_mask : int
        **Bit mask** (i.e., integer containing exactly one non-zero bit at the
        position uniquely identifying the desired bit) of the bit to be tested.

    Returns
    ----------
    bool
        ``True`` only if this bit in this bit field is on.
    '''

    # Return true only if this bit in this bit field is on. Dismantled, this
    # is:
    #
    # * "bit_field & bit_mask", either:
    #   * If this bit is on in this bit field, "bit_mask" and hence non-zero.
    #   * If this bit is off in this bit field, zero.
    # * "bool(...)", reducing this integer to a boolean.
    return bool(bit_field & bit_mask)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **reference** (i.e., pointer to the address of an object) facilities.
'''

# ....................{ IMPORTS                           }....................
import weakref
from betse.util.type.types import WeakRefProxyTypes, WeakRefTypes

# ....................{ PROXIERS                          }....................
def proxy_weak(obj: object) -> WeakRefProxyTypes:
    '''
    Weak reference to the passed object as a proxy object, such that the latter
    transparently masquerades as the former.

    The reference returned by this function raises a :data:`ReferenceError`
    exception on each attempt to access this reference *after* Python collects
    and hence deletes the passed object. If this object is:

    * Already a proxy object, this object is returned as is.
    * Callable, a :class:`CallableProxyType` object proxying this object is
      returned.
    * Uncallable, a :class:`ProxyType` object proxying this object is returned.

    Caveats
    ----------
    This function should *always* be called in lieu of the lower-level
    :func:`weakref.proxy` function, which unsafely raises the following
    exception when passed an existing proxy object:

        TypeError: cannot create weak reference to 'weakproxy' object

    This function should typically (but *not* necessarily always) be called in
    lieu of the lower-level :func:`get_weak` function, which silently and hence
    unsafely returns ``None`` rather than safely raising an exception after
    Python collects and hence deletes the passed object.

    Parameters
    ----------
    obj : object
        Object to be weakly proxied.

    Returns
    ----------
    WeakRefProxyTypes
        Weak reference to the passed object as a proxy object. When the passed
        object is garbage-collected, this proxy object raises a
        :data:`ReferenceError` exception rather than yielding the passed
        object. This proxy object is unhashable regardless of the hashability
        of the passed object, avoiding issues arising from the mutability of
        weak references and preventing their use as dictionary keys.

    See Also
    ---------
    :func:`get_weak`
        Lower-level getter returning ``None`` rather than raising an exception.
    '''

    # Return...
    return (
        # If this object is already a weak proxy, this object as is.
        obj if isinstance(obj, WeakRefProxyTypes) else
        # Else, a weak reference proxying this object.
        weakref.proxy(obj)
    )

# ....................{ REFERERS                          }....................
def refer_weak(obj: object) -> WeakRefTypes:
    '''
    Weak reference to the passed object as an unproxied callable object, such
    that calling the latter returns a strong (i.e., non-weak and hence
    standard) reference to the former if the former has yet to be collected
    *or* ``None`` otherwise (i.e., if the former has already been collected).

    This function explicitly supports common edge-cases unsupported by the
    underlying :func:`weakref.ref` function. Notably, if the passed object is:

    * Already a **weak reference** (i.e., object whose type is in the
      :class:`WeakRefTypes` tuple), that object is returned as is.
    * A **bound method** (i.e., function whose first parameter is *always*
      implicitly passed the same parent object), that method is wrapped with a
      weak reference specific to bound methods.
    * Any other object, that object is wrapped with a general-purpose unproxied
      weak reference.

    Caveats
    ----------
    This function should typically (but *not* necessarily always) be avoided in
    favour of the higher-level :func:`proxy_weak` function, which safely raises
    an exception rather than silently and hence unsafely returning ``None``
    after Python collects and hence deletes the passed object.

    This function should *always* be called in lieu of the lower-level
    :func:`weakref.ref` function, which unsafely raises the following
    exception when passed an existing weak reference:

        TypeError: cannot create weak reference to 'weakref' object

    Parameters
    ----------
    obj : object
        Object to be weakly refered to.

    Returns
    ----------
    WeakRefTypes
        Weak reference to the passed object as an unproxied callable object.
        When the passed object is garbage-collected, calling this weak
        reference returns ``None`` rather than the passed object. This weak
        reference is unhashable regardless of the hashability of the passed
        object, avoiding issues arising from the mutability of weak references
        and preventing their use as dictionary keys.

    Examples
    ----------
    Calling the weak reference created and returned by this function yields a
    strong reference to the object initially passed to this function. To avoid
    subtle race conditions between pure-Python code leveraging this weak
    reference and the garbage collector collecting the underlying object, this
    reference should *only* be called in the following usage pattern:

        >>> from betse.util.io.log import logs
        >>> from betse.util.io.log.conf import logconf
        >>> from betse.util.py import pyref
        >>> log_config_weak = pyref.refer_weak(logconf.get_log_conf())
        >>> log_config_strong = log_config_weak()
        >>> if log_config_strong is not None:
        ...     logs.log_debug('Verbose logging enabled: %r',
        ...                    log_config_strong.is_verbose)
        ... else:
        ...     logs.log_debug('Logging configuration no longer exists.')

    That is to say, the following recommendations should be observed:

    * Whenever this weak reference is called, the resulting strong reference
      should be persisted to a variable for subsequent reference. Conversely,
      this weak reference itself should *never* be directly referenced, as
      doing so invites race conditions with the garbage collector.
    * The first access of this strong reference should be embedded in an if
      conditional testing whether the underlying object:

      * Is still alive, in which case this reference is non-``None``.
      * Has already been collected, in which case this reference is ``None``.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.types import BoundMethodTypes

    # Return...
    return (
        # If this object is already a weak reference, this object as is.
        obj if isinstance(obj, WeakRefTypes) else
        # Else if this object is a bound method, a weak reference specific
        # to bound methods (whose obscure nature warrants special handling).
        weakref.WeakMethod(obj) if isinstance(obj, BoundMethodTypes) else
        # Else, a general-purpose weak reference referring to this object.
        weakref.ref(obj)
    )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Simulation configuration-specific **expression alias data descriptor** (i.e.,
object satisfying the data descriptor protocol aliasing a YAML-backed
simulation configuration option to a standard instance variable) functionality.
'''

# ....................{ IMPORTS                           }....................
from abc import ABCMeta
from betse.util.type.descriptor.expralias import expr_alias, expr_enum_alias
from betse.util.type.types import type_check

# ....................{ SUPERCLASSES                      }....................
class YamlAliasABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all simulation configuration-specific **expression
    alias data descriptor** (i.e., object satisfying the data descriptor
    protocol aliasing a YAML-backed simulation configuration option to a
    standard instance variable) subclasses.

    Motivation
    ----------
    The class principally exists for type validation -- namely, to
    differentiate the data descriptors returned by alias functions defined by
    this submodule (e.g., :func:`yaml_alias`) from those created by other
    means.
    '''

    pass

# ....................{ GLOBALS                           }....................
_YAML_ALIAS_BASE_CLASSES = (YamlAliasABC,)
'''
Tuple of all base classes of all simulation configuration-specific expression
alias data descriptors.

See Also
----------
:class:`YamlAliasABC`
    Further details.
'''

# ....................{ DESCRIPTORS                       }....................
@type_check
def yaml_alias(keys: str, *args, **kwargs) -> YamlAliasABC:
    '''
    Expression alias **data descriptor** (i.e., object satisfying the data
    descriptor protocol) specific to simulation configurations, dynamically
    aliasing a target variable bound to instances of the class instantiating
    this descriptor to a source Python expression performing one or more key
    lookups into the dictionary loaded from a YAML-formatted simulation
    configuration file.

    Parameters
    ----------
    keys : str
        Python expression evaluating to the value of an arbitrarily nested key
        of the dictionary loaded from the current simulation configuration,
        typically consisting of one or more ``[``- and ``]``-delimited key
        lookups into this same dictionary (e.g.,
        ``['variable settings']['noise']['dynamic noise']``).

    All remaining parameters are passed as is to the underlying
    :func:`expr_alias` function.

    Returns
    ----------
    YamlAliasABC
        Expression alias data descriptor as detailed above.

    See Also
    ----------
    :func:`expr_alias`
        Further details.
    '''

    return expr_alias(
        'self._conf' + keys, *args,
        base_classes=_YAML_ALIAS_BASE_CLASSES, **kwargs)


@type_check
def yaml_enum_alias(keys: str, *args, **kwargs) -> YamlAliasABC:
    '''
    Enumeration-specific expression alias **data descriptor** (i.e., object
    satisfying the data descriptor protocol) specific to simulation
    configurations, dynamically aliasing a target variable of the passed
    enumeration type bound to instances of the class instantiating this
    descriptor to an arbitrarily complex source Python expression performing
    one or more key lookups into the dictionary loaded from a YAML-formatted
    simulation configuration file.

    Parameters
    ----------
    keys : str
        Oner or more ``[``- and ``]``-delimited key lookups. See the
        :func:`yaml_alias` function for further details.

    All remaining parameters are passed as is to the underlying
    :func:`expr_enum_alias` function.

    Returns
    ----------
    YamlAliasABC
        Enumeration-specific expression alias data descriptor as detailed
        above.

    See Also
    ----------
    :func:`expr_enum_alias`
        Further details.
    '''

    return expr_enum_alias(
        'self._conf' + keys, *args,
        base_classes=_YAML_ALIAS_BASE_CLASSES, **kwargs)

# ....................{ DESCRIPTORS ~ predicate : float   }....................
def yaml_alias_float_positive(keys: str) -> YamlAliasABC:
    '''
    Simulation configuration expression alias data descriptor, dynamically
    aliasing a target floating point variable with values constrained to be
    **positive** (i.e., strictly greater than 0.0).

    See Also
    ----------
    :func:`yaml_alias`
        Further details.
    '''

    return yaml_alias(
        keys=keys,
        cls=float,
        predicate_expr='value > 0.0',
        predicate_label='positive',
    )


def yaml_alias_float_nonnegative(keys: str) -> YamlAliasABC:
    '''
    Simulation configuration expression alias data descriptor, dynamically
    aliasing a target floating point variable with values constrained to be
    **non-negative** (i.e., greater than or equal to 0.0).

    See Also
    ----------
    :func:`yaml_alias`
        Further details.
    '''

    return yaml_alias(
        keys=keys,
        cls=float,
        predicate_expr='value >= 0.0',
        predicate_label='non-negative',
    )


def yaml_alias_float_percent(keys: str) -> YamlAliasABC:
    '''
    Simulation configuration expression alias data descriptor, dynamically
    aliasing a target floating point variable with values constrained to be
    **percentages** (i.e., in the range ``[0.0, 100.0]``).

    See Also
    ----------
    :func:`yaml_alias`
        Further details.
    '''

    return yaml_alias(
        keys=keys,
        cls=float,
        predicate_expr='0.0 <= value <= 100.0',
        predicate_label='percentage',
    )

# ....................{ DESCRIPTORS ~ predicate : int     }....................
def yaml_alias_int_positive(keys: str) -> YamlAliasABC:
    '''
    Simulation configuration expression alias data descriptor, dynamically
    aliasing a target integer variable with values constrained to be
    **positive** (i.e., strictly greater than 0).

    See Also
    ----------
    :func:`yaml_alias`
        Further details.
    '''

    return yaml_alias(
        keys=keys,
        cls=int,
        predicate_expr='value > 0',
        predicate_label='positive',
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests exercising memoization decorators.
'''

# ....................{ IMPORTS                            }....................
# import pytest

# ....................{ TESTS                              }....................
def test_callable_cached() -> None:
    '''
    Unit test the :func:`betse.util.type.callables.func_cached` decorator.
    '''

    # Defer heavyweight imports.
    from betse.util.type.decorator.decmemo import func_cached

    # Class containing the callable to be cached.
    class Dreamland(object):
        def __init__(self, dreams: int) -> None:
            self.dreams = dreams

        @func_cached
        def dreams_cached(self) -> int:
            # Instance variable to be both cached and returned.
            dreams_cached = self.dreams * 2

            # To detect erroneous attempts to call this callable multiple
            # times for the same object, modify the instance variable producing
            # this return value in a predictable way on each call.
            self.dreams /= 2

            # Return this variable's value.
            return dreams_cached

    # Value of the "Keeper.dreams" attribute *BEFORE* invoking dreams_cached().
    DREAM_COUNT_PRECACHED = 7

    # Value of the "Keeper.dreams" attribute *AFTER* invoking dreams_cached().
    DREAM_COUNT_POSTCACHED = DREAM_COUNT_PRECACHED / 2

    # Value of the "Keeper.dreams_cached" callable.
    DREAM_COUNT_CACHED = DREAM_COUNT_PRECACHED * 2

    # Instance of this class initialized with this value.
    rebellion = Dreamland(dreams=DREAM_COUNT_PRECACHED)

    # Assert this attribute to be as initialized.
    assert rebellion.dreams == DREAM_COUNT_PRECACHED

    # Assert this callable to be as cached.
    assert rebellion.dreams_cached() == DREAM_COUNT_CACHED

    # Assert this attribute to have been modified by this callable call.
    assert rebellion.dreams == DREAM_COUNT_POSTCACHED

    # Assert this callable to still be as cached.
    assert rebellion.dreams_cached() == DREAM_COUNT_CACHED

    # Assert this attribute to *NOT* have been modified again, validating that
    # the prior callable access returned the previously cached value rather than
    # recalling this callable method.
    assert rebellion.dreams == DREAM_COUNT_POSTCACHED


def test_property_cached() -> None:
    '''
    Unit test the :func:`betse.util.type.callables.property_cached` decorator.
    '''

    # Defer heavyweight imports.
    from betse.util.type.decorator.decmemo import property_cached

    # Class containing the property to be cached.
    class Keeper(object):
        def __init__(self, keys: int) -> None:
            self.keys = keys

        @property_cached
        def keys_cached(self) -> int:
            # Property value to be both cached and returned.
            keys_cached = self.keys * 2

            # To detect erroneous attempts to call this property method multiple
            # times for the same object, modify the object variable from which
            # this property value derived in a predictable way on each call of
            # this property method.
            self.keys /= 2

            # Return this property value.
            return keys_cached

    # Value of the "Keeper.keys" attribute *BEFORE* invoking keys_cached().
    KEY_COUNT_PRECACHED = 7

    # Value of the "Keeper.keys" attribute *AFTER* invoking keys_cached().
    KEY_COUNT_POSTCACHED = KEY_COUNT_PRECACHED / 2

    # Value of the "Keeper.keys_cached" property.
    KEY_COUNT_CACHED = KEY_COUNT_PRECACHED * 2

    # Instance of this class initialized with this value.
    i_want_out = Keeper(keys=KEY_COUNT_PRECACHED)

    # Assert this attribute to be as initialized.
    assert i_want_out.keys == KEY_COUNT_PRECACHED

    # Assert this property to be as cached.
    assert i_want_out.keys_cached == KEY_COUNT_CACHED

    # Assert this attribute to have been modified by this property call.
    assert i_want_out.keys == KEY_COUNT_POSTCACHED

    # Assert this property to still be as cached.
    assert i_want_out.keys_cached == KEY_COUNT_CACHED

    # Assert this attribute to *NOT* have been modified again, validating that
    # the prior property access returned the previously cached value rather than
    # recalling this property method.
    assert i_want_out.keys == KEY_COUNT_POSTCACHED
